# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Domain-specific language for writing numeric tests.

This module provides a few classes and functions to write numeric tests.

Let us import the relevant modules first:

    >>> from collections import OrderedDict
    >>> from valjean.eponine.dataset import Dataset
    >>> import numpy as np

Now we create a toy data set:

    >>> x = np.linspace(-5., 5., num=100)
    >>> y = x**2
    >>> error = np.zeros_like(y)
    >>> bins = OrderedDict()
    >>> bins['x'] = x
    >>> parabola = Dataset(y, error, bins=bins)

We perturb the data by applying some small amount of noise:

    >>> eps = 1e-8
    >>> noise = np.random.uniform(-eps, eps, parabola.shape)
    >>> y2 = y + noise
    >>> parabola2 = Dataset(y2, error, bins=bins)

Now we can test if the new dataset is equal to the original one:

    >>> from valjean.gavroche.test import TestEqual
    >>> test_equality = TestEqual(parabola, parabola2, name="parabola",
    ...                           description="equality test")
    >>> test_equality_res = test_equality.evaluate()
    >>> print(bool(test_equality_res))
    False

However, they are approximately equal:

    >>> from valjean.gavroche.test import TestApproxEqual
    >>> test_approx = TestApproxEqual(parabola, parabola2, name="parabola",
    ...                               description="approx equal test")
    >>> test_approx_res = test_approx.evaluate()
    >>> print(bool(test_approx_res))
    True
'''
from abc import ABC, abstractmethod
import numpy as np


class CheckBinsException(Exception):
    '''An error is raised when check bins fails.'''


def same_arrays(arr1, arr2):
    '''Return `True` if `arr1` and `arr2` are equal.

    :param arr1: the first array.
    :param arr2: the second array.
    '''
    return np.array_equal(arr1, arr2)


def same_bins(bins1, bins2):
    '''Return `True` if all the coordinate arrays are compatible.

    :param bins1: the first dictionary of coordinate arrays.
    :param bins2: the second dictionary of coordinate arrays.
    '''
    if bins1.keys() != bins2.keys():
        return False
    return all(same_arrays(bins1[k], bins2[k]) for k in bins1.keys())


def same_bins_datasets(*datasets):
    '''Return `True` if all datasets have the same coordinates.

    :param datasets: any number of datasets.
    :type datasets: :class:`~valjean.eponine.dataset.Dataset`
    '''
    for dataset in datasets[1:]:
        if not same_bins(datasets[0].bins, dataset.bins):
            return False
    return True


def check_bins(*datasets):
    '''Check if the datasets have compatible coordinates, raise if not.

    :raises ValueError: if the datasets do not have compatible coordinates.
    '''
    if not same_bins_datasets(*datasets):
        datasets_str = '\n'.join(str(dat) for dat in datasets)
        raise CheckBinsException(f'Inconsistent coordinates: \n{datasets_str}')


class Test(ABC):
    '''Generic class for comparing any kind of results.

    Base class for tests.
    '''

    def __init__(self, *, name, description='', labels=None):
        '''Initialize the :class:`~.Test` object with a name, a description of
        the test (may be long) and labels if needed.

        The test is actually performed in the :meth:`evaluate` method, which is
        abstract in the base class and must be implemented by sub-classes.

        :param str name: name of the test, this string will typically end up in
                         the test report as a section name.
        :param str description: description of the test exepcted with context,
                                this string will typically end up in the test
                                report.
        :param dict labels: labels to be used for test classification in
                            reports (for example category, input file name,
                            type of result, ...)
        '''
        self.name = name
        self.description = description
        self.labels = {} if labels is None else labels.copy()

    @abstractmethod
    def evaluate(self):
        '''Evaluate the test on the given inputs.

        Must return a subclass of :class:`~.TestResult`.
        '''

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield self.__class__.__name__.encode('utf-8')
        yield self.name.encode('utf-8')
        yield self.description.encode('utf-8')
        # labels intentionally excluded; this makes it possible to put any type
        # of items in the label values, and not just strings

    # tell pytest that this class and derived classes should NOT be collected
    # as tests
    __test__ = False


class TestDataset(Test):
    '''Generic class for comparing datasets.'''

    def __init__(self, dsref, *datasets, name, description='', labels=None):
        '''Initialisation of :class:`~.TestEqual`:

        :param str name: name of the test (in analysis)
        :param str description: specific description of the test
        :param dict labels: labels to be used for test classification in
                            reports (for example category, input file name,
                            type of result, ...)
        :param Dataset dsref: reference dataset
        :param list(Dataset) datasets: list of datasets to be compared to
                                       reference dataset
        '''
        super().__init__(name=name, description=description, labels=labels)
        self.dsref = dsref
        self.datasets = datasets
        if not datasets:
            raise ValueError('At least one dataset expected to be compared to '
                             'the reference one')

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        yield from self.dsref.data()
        for dataset in self.datasets:
            yield from dataset.data()

    @abstractmethod
    def evaluate(self):
        '''Evaluate the test on the given datasets.

        Must return a subclass of :class:`~.TestResult`.
        '''


class TestResult(ABC):
    '''Base class for test results.

    This result should be filled by :class:`~.Test` daughter classes.
    '''

    def __init__(self, test):
        '''Initialisation of :class:`~.TestResult`.

        :param test: the used test
        :type test: :class:`~.Test` used
        '''
        self.test = test

    @abstractmethod
    def __bool__(self):
        pass

    # tell pytest that this class and derived classes should NOT be collected
    # as tests
    __test__ = False


class TestResultFailed(TestResult):
    '''Class for failed TestResults when an exception was raised during the
    evaluation.
    '''

    def __init__(self, test, msg):
        '''Initialisation of :class:`~.TestResult`.

        :param test: the used test
        :type test: :class:`~.Test` used
        '''
        super().__init__(test)
        self.msg = msg

    def __bool__(self):
        return False


class TestResultEqual(TestResult):
    '''Result from :class:`TestEqual`.'''

    def __init__(self, test, equal):
        '''Initialisation of the result from :class:`~.TestEqual`:

        :param test: the used test
        :type test: :class:`~.TestEqual`
        :param equal: result from the test
        :type equal: :class:`list` (``numpy.bool_``) if datasets are
            :obj:`numpy.generic`, :class:`list` (:obj:`numpy.ndarray`) if
            datasets are :obj:`numpy.ndarray` with ``dtype == bool``.
        '''
        super().__init__(test)
        self.equal = equal

    def __bool__(self):
        '''Return the result of the test: ``True`` or ``False`` or raises an
        exception when it is not suitable.'''
        return bool(np.all(self.equal))


class TestEqual(TestDataset):
    '''Test if the datasets values are equal. Errors are ignored.'''

    def evaluate(self):
        '''Evaluation of :class:`~.TestEqual`.

        :returns: :class:`~.TestResultEqual`
        '''
        equal = []
        for _ds in self.datasets:
            check_bins(self.dsref, _ds)
            equal.append(np.equal(self.dsref.value, _ds.value))
        return TestResultEqual(self, equal)

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')


class TestResultApproxEqual(TestResult):
    '''Result from :class:`TestApproxEqual`.'''

    def __init__(self, test, approx_equal):
        '''Initialisation of the result from :class:`~.TestApproxEqual`:

        :param test: the used test
        :type test: :class:`~.TestApproxEqual`
        :param approx_equal: result from the test
        :type approx_equal: :obj:`numpy.generic` if datasets are
                            :obj:`numpy.generic`, :obj:`numpy.ndarray` if
                            datasets are :obj:`numpy.ndarray`.
                            In both cases ``dtype == bool``.
        '''
        super().__init__(test)
        self.approx_equal = approx_equal

    def __bool__(self):
        return bool(np.all(self.approx_equal))


class TestApproxEqual(TestDataset):
    '''Test if the datasets values are equal within the given tolerances.
    Errors are ignored.
    '''

    def __init__(self, dsref, *datasets, name, description='', labels=None,
                 rtol=1e-5, atol=1e-8):
        # pylint: disable=too-many-arguments
        '''Initialisation of :class:`~.TestApproxEqual`:

        :param str name: local name of the test
        :param str description: specific description of the test
        :param dict labels: labels to be used for test classification in
                            reports (for example category, input file name,
                            type of result, ...)
        :param Dataset dsref: reference dataset
        :param list(Dataset) datasets: list of datasets to be compared to
                                       reference dataset
        :param float rtol: relative tolerance, default = :math:`10^{-5}`
        :param float atol: absolute tolerance, default = :math:`10^{-8}`

        To get more details on `rtol` and `atol` parameters, see
        :func:`numpy.isclose`.
        '''
        super().__init__(dsref, *datasets,
                         name=name, description=description, labels=labels)
        self.rtol = rtol
        self.atol = atol

    def evaluate(self):
        '''Evaluation of :class:`~.TestApproxEqual`.

        :returns: :class:`~.TestResultApproxEqual`
        '''
        approx_equal = []
        for _ds in self.datasets:
            check_bins(self.dsref, _ds)
            approx_equal.append(np.isclose(self.dsref.value, _ds.value,
                                           rtol=self.rtol, atol=self.atol))
        return TestResultApproxEqual(self, approx_equal)

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        yield float(self.rtol).hex().encode('utf-8')
        yield float(self.atol).hex().encode('utf-8')



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Main :program:`valjean` executable.'''

import sys
import argparse
import pkgutil
import importlib
import logging

from . import commands
from .. import LOG_FILE_FORMAT, __version__
from ..config import Config
from ..path import ensure


LOGGER = logging.getLogger(__name__)


def main(argv=None):
    '''Main entry point for the :program:`valjean` executable.'''
    if argv is None:
        argv = sys.argv[1:]
    LOGGER.debug('arguments: %s', argv)

    parser = make_parser()
    args = parser.parse_args(argv)
    LOGGER.debug('parsed args: %s', args)

    config = process_options(args)

    if hasattr(args, 'func'):
        # collect the tasks and call the processing function
        args.func(args, config)
    else:
        parser.print_help()


def make_parser():
    '''Construct the argument parser.'''
    parser = argparse.ArgumentParser(
        prog='valjean',
        description='VALidation, Journal d’Évolution et ANalyse',
        epilog='You can pass the -h/--help option to the subcommands to get '
        'specific help.')

    # -V/--version
    version_string = '%(prog)s ' + __version__
    parser.add_argument('-V', '--version', action='version',
                        version=version_string,
                        help='show the code version number and exit')
    parser.add_argument('-v', '--verbose', action='count',
                        help='increase verbosity')
    parser.add_argument('-c', '--config', action='store',
                        help='use the specified configuration file')
    parser.add_argument('-l', '--log', help='path to the log file')

    # here come the subcommands
    cmd_parsers = parser.add_subparsers(title='Valid commands',
                                        dest='command_name')

    # import each submodule in commands.* and create a new subparser for it
    prefix = commands.__name__ + '.'
    submods_iter = pkgutil.iter_modules(commands.__path__)
    cmd_objs = []
    for _, modname, _ in submods_iter:
        cmd_name = modname.capitalize() + 'Command'
        LOGGER.debug('registering new command %s from class %s',
                     modname, cmd_name)
        module = importlib.import_module(prefix + modname)
        cmd_cls = getattr(module, cmd_name)
        cmd_obj = cmd_cls()
        cmd_objs.append(cmd_obj)

    for cmd_obj in cmd_objs:
        cmd_parser = cmd_parsers.add_parser(cmd_obj.NAME,
                                            help=cmd_obj.HELP,
                                            description=cmd_obj.DESC,
                                            aliases=cmd_obj.ALIASES)
        cmd_obj.register(cmd_parser)

    return parser


def process_options(args):
    '''Process the parsed options.

    :returns: The configuration.'''

    # configuration file
    if args.config is not None:
        config = Config.from_file(args.config)
    else:
        config = Config()

    # verbosity
    if args.verbose is not None:
        if args.verbose >= 1:
            log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    LOGGER.setLevel(log_level)
    for handler in LOGGER.handlers:
        handler.setLevel(log_level)

    # log to file
    if args.log is not None:
        formatter = logging.Formatter(LOG_FILE_FORMAT,
                                      datefmt='%Y-%m-%d %H:%M:%S')
        ensure(args.log)
        handler = logging.FileHandler(args.log)
        handler.setFormatter(formatter)
        handler.setLevel(log_level)
        LOGGER.addHandler(handler)

    return config



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Module performing a scan of Tripoli-4 output listing in order to only keep
relevant parts of it = results to be used for V&V or analysis or the run.

Summary
-------

* Quickly reads the results file
* Recognize beginning and end of results sections
* Get the required number of batchs
* Get the edition batch numbers (if exists)


Use of :mod:`~valjean.eponine.tripoli4.scan`
-----------------------------------------------

.. doctest:: scan
    :hide:

    >>> import os
    >>> work_dir = 'scan'
    >>> os.mkdir(work_dir)
    >>> with open(os.path.join(work_dir, 'spam.res'), 'w') as tmpfile:
    ...     print("BATCH 10\\n"
    ...           "initialization time (s): 7\\n"
    ...           " batch number : 10\\n"
    ...           "RESULTS ARE GIVEN FOR SOURCE INTENSITY : 1.000000e+00\\n"
    ...           "Edition after batch number : 10\\n"
    ...           "simulation time (s) : 1\\n"
    ...           "NORMAL COMPLETION",
    ...           file=tmpfile)

To use this module you need to create a :class:`Scanner` object giving at least
the path to the file you want to read. This file should be a Tripoli-4 output
containing at least the flags precised in :ref:`eponine-t4-scan-caveats`.

    >>> import os
    >>> from valjean.eponine.tripoli4.scan import Scanner
    >>> results = Scanner(os.path.join(work_dir, 'spam.res'))
    >>> results.normalend
    True
    >>> len(results)
    1
    >>> results.times['initialization_time']
    7

    The expected key of the :class:`Scanner` object is the batch number, not an
    index. If you have the index you need to obtain the corresponding batch
    number first.

    >>> 'simulation time' in results[10]  # batch number = 10
    True
    >>> 'simulation time' in results[-1]  # -1 can only be an index, like 0
    Traceback (most recent call last):
        ...
    KeyError: -1
    >>> 'simulation time' in results[results.batch_number(-1)]
    True

.. note::

   It will probably be better to directly load a test file...


.. _eponine-t4-scan-caveats:

Caveats
-------

Beginning and end of results sections
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Important for the scan: results will be kept

* **from** "RESULTS ARE GIVEN"
* **to** an end flag available in the list ``Scanner.end_flags``.
  Possibilities are:

  * Default end flag is ``"simulation time"``;
  * For exploitation jobs use ``"exploitation time"`` (example: Green bands);
  * for jobs running in parallel, ``"elapsed time"`` will also appear, after
    ``"simulation time"`` normal.

Module API
----------
'''
import logging
from collections.abc import Mapping
from collections import OrderedDict


LOGGER = logging.getLogger(__name__)

# get profile from globals (cleaner)
if 'profile' not in globals()['__builtins__']:
    def profile(fmem):
        '''Deactivate profiling if not required in command line.'''
        return fmem


class PhEmEpBalanceOutput:
    '''Class to store photon-electron-positron balance.'''

    def __init__(self):
        self.content = []
        self.in_phemep = False
        self.count = 0

    def _count(self, line):
        if "Total" in line:
            self.count += 1
            if self.count == 3:
                self.in_phemep = False

    def add_line(self, line):
        '''Add line to the photon electron positron output and counts.'''
        self.content.append(line)
        self._count(line)


class HomogMatOutput:
    '''Class to store the homogenized material output.'''

    def __init__(self):
        self.content = []
        self.in_dump = False
        self.nb_groups = 0
        self.counting = False
        self.nb_corr_lines = 0

    def _count(self, line):
        '''Count the number of bins for the homogenized material dump.'''
        if self.counting:
            if self.nb_corr_lines > 0:
                self.nb_corr_lines += 1
            elif line[0].isdigit():
                self.nb_groups += 1
        if 'dump total section :' in line:
            self.counting = True
        elif 'dump absorption section :' in line:
            self.counting = False
        elif "correlation between absorption and total cross section" in line:
            self.nb_corr_lines += 1
            self.counting = True
        elif self.nb_corr_lines > self.nb_groups:
            self.counting = False
            self.in_dump = False

    def add_line(self, line):
        '''Add line to the homogenized material output and count.'''
        self.content.append(line)
        self._count(line)


class BatchResultScanner:
    # pylint: disable=too-many-instance-attributes
    '''Class to build batchs collection.

    :param int current_batch: current batch number
    :param bool para: flag to identify outputs run in parallel
    :param str line: current line in file

    .. note::

      * **current_batch** normally refers to lines before flag
        "RESULTS ARE GIVEN". They can be used if "Edition after batch number"
        does not appear in the file.
      * **line** should contain "RESULTS ARE GIVEN" here. It is used to
        initialize the list of strings corresponding to the result block.

    '''

    def __init__(self, current_batch, para, line):
        self.batch_counts = {'number': -1,
                             'current': current_batch,
                             'greater': 0}
        self.result = [line]
        self.para = para
        self.phemep_balance = PhEmEpBalanceOutput()
        self.homog_mat = HomogMatOutput()

    def build_result(self, line):
        '''Scan line to build batch result: mainly deals with specific patterns
        and store line.

        :param str line: last line to be taken into account
        '''
        if "Edition after batch number" in line:
            self.batch_counts['number'] = int(line.split()[-1])
        elif '#'*64 in line and not self.phemep_balance.in_phemep:
            LOGGER.debug('phe+e- balance')
            self.phemep_balance.in_phemep = True
        elif "DUMP HOMOGENIZED MATERIAL" in line:
            LOGGER.debug("In DUMP HOMOGENIZED MATERIAL")
            self.homog_mat.in_dump = True
        if self.para and "number of batches used" in line:
            self._set_greater_batch_number(line)
        self._store_line(line)

    def _store_line(self, line):
        if self.phemep_balance.in_phemep:
            self.phemep_balance.add_line(line)
        elif self.homog_mat.in_dump:
            self.homog_mat.add_line(line)
        else:
            self.result.append(line)

    def _set_greater_batch_number(self, line):
        newbatch = int(line.split()[4])
        if self.batch_counts['greater'] < newbatch:
            self.batch_counts['greater'] = newbatch

    def check_batch_number(self):
        '''Check batch number value and replace it by current value or by the
        greater value if needed.
        '''
        if not self.para:
            if self.batch_counts['number'] != self.batch_counts['current']:
                LOGGER.info("Edition batch (%d) different from "
                            "current batch (%d)",
                            self.batch_counts['number'],
                            self.batch_counts['current'])
                LOGGER.info("If no Edition batch keep current batch, "
                            "else keep edition batch")
                if self.batch_counts['number'] < self.batch_counts['current']:
                    self.batch_counts['number'] = self.batch_counts['current']
            if self.batch_counts['current'] == 0:
                LOGGER.warning("Current batch = 0, something to check ?")
        else:
            if self.batch_counts['greater'] > self.batch_counts['number']:
                self.batch_counts['number'] = self.batch_counts['greater']

    def get_result(self):
        '''Send result.
        Called if end flag has been found, add last line and concatenates
        result.

        :return: string build from list of strings junction
        '''
        LOGGER.debug("END FLAG found, batch number = %d, "
                     "current batch = %d, greater batch = %d",
                     self.batch_counts['number'],
                     self.batch_counts['current'],
                     self.batch_counts['greater'])
        return ''.join(self.result)


class ScannerException(Exception):
    '''An error that may be raised by the :class:`Scanner` class.'''


class Scanner(Mapping):
    # pylint: disable=too-many-instance-attributes
    '''Class to scan the Tripoli-4 listing and keep the relevant parts of it
    like results per batch used for edition or times.

    There are no class variables, but instance variables (initialized when
    the object is built or when the file is read). They are directly accessible
    from the object. Main results are accessible directly from the
    :obj:`Scanner` object.

    **Instance variables:**

    `fname` (:class:`str`)
        name of the file that will be scanned

    `batches` (:class:`dict`)
        keep the number of batches and the packet_length
        (read from file `fname`)

    `normalend` (:class:`bool`)
        presence of "NORMAL COMPLETION"

    `end_flags` (:class:`list` (:class:`str`))
        possible end flags to stop the saving of results

    `para` (:class:`bool`)
        `True` in parallel mode

    `countwarnings` (:class:`int`)
        count number of warnings (for statistics)

    `counterrors` (:class:`int`)
        count number of errors (for statistics)

    `times` (:class:`collections.OrderedDict`)
        save times (initialization, simulation, exploitation and elapsed if
        exists). Mandatory ones are ``'initialization time'`` and ``'simulation
        time'`` or ``'exploitation time'``. ``'elapsed time'`` only appears in
        listings from parallel jobs.

    `last_generator_state` (:class:`str`)
        keep the random generator state (not included in the result as given
        after `endflag`)

    `phemep_balance` (:class:`dict`)
        keep the photon electron positron balance as :class:`str` indexed by
        batch number

    `homog_mat` (:class:`dict`)
        keep the homogenized material dump as :class:`str` indexed by batch
        number


    **Available methods:**

    :class:`Scanner` inherits from :class:`collections.abc.Mapping` so many
    methods are implemented or available by default: ``keys``, ``items``,
    ``values``, ``get``, ``__contains__`` (used via ``in``). ``__getitem``
    (used with ``[]``), ``__iter__`` (when iterators are required), ``__len__``
    and ``__reversed__`` are redefined.

    This class points on the :obj:`collections.OrderedDict` that it contains.
    The keys are the available batch numbers in the Tripoli-4 output, obtained
    using ``scan_obj.keys()``.
    '''

    @profile
    def __init__(self, fname, end_flag=""):
        '''Initialize the instance from the file `fname`, meaning reads the
        file and store the relevant parts of it, i.e. result block for each
        batch edition.

        Results are stored in an internal :obj:`collections.OrderedDict`:
        ``{batch_number_1: 'result_1', batch_number_2: 'result_2', ...}``

        * ``batch_number_*`` is `int`
        * ``result_*`` is `str`
        * Order follows the listing order, so increasing ``batch_number``

        Members needed at initialization:

        :param str fname: name of the input file
        :param str end_flag: end flag of the results block in Tripoli-4 listing
        '''
        self.fname = fname
        self.batches = {"batches": None,
                        "packet_length": 1}
        self.tasks = 1
        self.normalend = False
        self.end_flags = ["simulation time", "exploitation time",
                          "elapsed time"]
        if end_flag:
            self.end_flags.insert(0, end_flag)
        self.para = False
        self.partial = False
        self.countwarnings = 0
        self.counterrors = 0
        self.times = OrderedDict()
        self.last_generator_state = ""
        self.phemep_balance = {}
        self.homog_mat = {}
        self._fatal_error = []
        self._collres = OrderedDict()
        self._get_collres()

    def _check_input_data(self, line):
        '''Get some parameters from introdcution of the results file, i.e.
        from the data file. Typically the number of batches required.
        '''
        if "BATCH" in line and '_' not in line and "THIS" not in line:
            indbatch = line.split().index('BATCH')
            self.batches["batches"] = (int(line.split()[indbatch+1])
                                       if len(line.split()) > 1
                                       else None)
        elif "number of tasks is" in line:
            self.para = True
            self.tasks = int(line.split()[5])
        elif "BATCH_PER_SIMULATOR" in line:
            self.batches["batches"] = (int(line.split()[1])
                                       * (self.tasks-2
                                          if self.tasks > 1 else 1))
        elif "PACKET_LENGTH" in line:
            LOGGER.info("Batchs grouped by packets -> "
                        "number of batchs expected divided "
                        "by PACKET_LENGTH in PARA")
            indpacket = line.split().index('PACKET_LENGTH')
            if self.para:
                self.batches["packet_length"] = int(line.split()[indpacket+1])
            LOGGER.debug("Packet length = %d", self.batches["packet_length"])
        elif "initialization time" in line:  # correspond to end of data file
            self.times['initialization_time'] = int(line.split()[3])

    def _is_end_flag(self, line):
        # len(end_flags) == 3: default list, no user's end flag added
        if "time" not in line and len(self.end_flags) == 3:
            return None
        for end_flag in self.end_flags:
            if end_flag in line:
                return end_flag
        return None

    def _add_time(self, end_flag, line):
        batch_number = next(reversed(self._collres)) if self._collres else 0
        eflag = end_flag.replace(' ', '_')
        time = (int(line.split()[-1]) if line.split()[-1].isdigit()
                else "Not a time")
        self.times.setdefault(eflag, {}).setdefault(batch_number, time)
        if self.times[eflag][batch_number] != time and 'time' in end_flag:
            if self.partial:
                LOGGER.warning("Partial edition: %s set to last found", eflag)
                self.times[eflag][batch_number] = time
            else:
                LOGGER.warning("More than one %s found for the same batch, %s",
                               eflag, batch_number)

    def _set_counters_and_flags(self, line):
        if "WARNING" in line:
            self.countwarnings += 1
        elif "ERROR" in line:
            self.counterrors += 1
            if "FATAL ERROR" in line:
                self._fatal_error = ["fatal error in Tripoli-4:\n\n"]
        elif "PARTIAL EDITION" in line:
            self.partial = True
        elif "NORMAL COMPLETION" in line:
            self.normalend = True

    def _additional_outputs(self, batch_scan):
        bnum = batch_scan.batch_counts['number']
        if batch_scan.phemep_balance.count == 3:
            self.phemep_balance[bnum] = ''.join(
                batch_scan.phemep_balance.content)
        if (batch_scan.homog_mat.nb_corr_lines != 0
                and not batch_scan.homog_mat.in_dump):
            self.homog_mat[bnum] = ''.join(batch_scan.homog_mat.content)

    @profile
    def _get_collres(self):
        # pylint: disable=too-many-branches
        '''Read the file and store all relevant information.
        '''
        generator_state = []
        current_batch = 0
        _batch_scan = None
        with open(self.fname, errors='ignore', encoding='utf-8') as fil:
            for line in fil:
                if line.lstrip().startswith("//"):  # comment in the jdd
                    continue
                if line.lstrip().startswith("!!!"):
                    continue
                self._set_counters_and_flags(line)
                if _batch_scan:
                    _batch_scan.build_result(line)
                    end_flag = self._is_end_flag(line)
                    if end_flag:
                        LOGGER.debug('end flag "%s" found', end_flag)
                        # check batch number has to be before get result to
                        # modify batch number before storage if necessary...
                        _batch_scan.check_batch_number()
                        batch_number = _batch_scan.batch_counts['number']
                        self._collres[batch_number] = _batch_scan.get_result()
                        self._additional_outputs(_batch_scan)
                        _batch_scan = None
                        # ordered dictionary -> unique keys, so only last kept
                        self._add_time(end_flag, line)
                        continue
                elif self._fatal_error:
                    self._fatal_error.append(line)
                elif not self.times:
                    self._check_input_data(line)
                    continue
                elif "RESULTS ARE GIVEN" in line:
                    _batch_scan = BatchResultScanner(current_batch, self.para,
                                                     line)
                elif (self.partial and line.startswith(' number of batch')
                      or line.startswith(' batch number :')):
                    current_batch = int(line.split()[-1])
                elif self._is_end_flag(line):
                    check_current_batch = (
                        current_batch == list(self._collres.keys())[-1]
                        if self._collres else False)
                    if not self.para and not check_current_batch:
                        continue
                    # still needed to be sure "elapsed time" appears in
                    # parallel jobs
                    self._add_time(self._is_end_flag(line), line)
                elif ("Type and parameters of random generator "
                      "at the end of simulation:" in line):
                    generator_state.append('')
                    continue
                elif generator_state:
                    generator_state.append(line)
                    if "COUNTER" in line:
                        self.last_generator_state = ''.join(generator_state)
                        generator_state = []
        if not self._collres and _batch_scan:
            raise ScannerException("No scan result built: "
                                   "no end flag found in the file")

    def __getitem__(self, batch_number):
        '''Get result corresponding to batch_number.

        A warning is printed if the last batch_number doesn't correspond to
        the number of batchs required.

        :param int batch_number: batch number (>0), corresponding the keys of
            Scanner.
        :raises KeyError: if ``batch_number`` does not exist (for example if
            confusion between ``batch_number`` and ``batch_index`` using -1 or
            0)

        Use: ``Scanner[X]``
        '''
        LOGGER.debug("__getitem__, batch number = %d", batch_number)
        try:
            return self._collres[batch_number]
        except KeyError:
            LOGGER.error("Wrong batch number required, batch number %s "
                         "doesn't exist, please change it to an existing "
                         "one.", batch_number)
            raise

    def __iter__(self):
        '''Iteration over the collection of results, on the keys to match
        :obj:`dict` and :obj:`collections.OrderedDict` behaviour.
        '''
        yield from self._collres.__iter__()

    def __len__(self):
        '''Return length of the collection of results, equivalent to get the
        number of edited batchs.
        '''
        return len(self._collres)

    def __reversed__(self):
        '''Reversed the :obj:`collections.OrderedDict` order (easier to get
        last element).
        '''
        yield from self._collres.__reversed__()

    def batch_index(self, batch_number):
        '''Get the index of the batch_number in the results list.'''
        return list(self._collres.keys()).index(batch_number)

    def batch_number(self, batch_index):
        '''Get the batch number from the batch index.'''
        return list(self.keys())[batch_index]

    def fatal_error(self):
        '''Return the fatal error message if found.'''
        return ''.join(self._fatal_error)

    def global_variables(self, batch_number):
        '''Return a dictionary of the global quantities in the TRIPOILI-4
        output:

        * warnings
        * errors
        * number of tasks (to distinguish MONO and PARA for example)
        * normal end
        * required batches
        * partial (if the job as been stopped)
        * batch number, especially when 'edition after batch number' is not in
          the Tripoli-4 output
        * t4_file: path to the scanned file
        * times of the required batch (can be simulation time, initialisation
          time, elapsed time)

        :param int batch_number: batch number (used for times)
        :rtype: dict
        '''
        gvars = {'warnings': self.countwarnings,
                 'errors': self.counterrors,
                 'number_of_tasks': self.tasks,
                 'normal_end': self.normalend,
                 'required_batches': (self.batches["batches"]
                                      // self.batches["packet_length"]
                                      if self.batches["batches"] is not None
                                      else None),
                 'partial': self.partial,
                 'batch_number': batch_number,
                 't4_file': self.fname}
        for ktime, val in self.times.items():
            gvars[ktime] = (val if isinstance(val, int)
                            else val.get(batch_number, None))
        return gvars

    def print_statistics(self):
        '''Print statistics of the listing scanned: normal end, number of
        warnings and errors.
        '''
        print("Normal end of the jdd:", self.normalend)
        print("Number of warnings found:", self.countwarnings)
        print("Number of errors found:", self.counterrors)

    def check_times(self):
        '''Check times.

        If the results come from a parallel job, the *times* dictionary should
        contain the ``'elapsed_time'`` key.
        In all kind of outputs or ``'simulation_time'`` or
        ``'exploitation_time'`` should be found (in parallel case the only
        possibility is in reality ``'simulation_time'``).
        Finally, while ``'initialization_time'`` value is an *int* as occuring
        only once, the others are lists with length equal to the number of
        batches in the edition for ``'simulation_time'`` and
        ``'exploitation_time'`` and equal to the number of batches + 1 for
        ``'elapsed_time'``.
        If all these checks are successful True is returned, else False.

        :rtype: bool
        '''
        if self.para:
            if 'elapsed_time' not in self.times:
                return False
            if self.partial:
                return True
        setimes = {'simulation_time', 'exploitation_time'}
        if not setimes.intersection(self.times):
            return False
        for time, times in self.times.items():
            if time in setimes and len(times) != len(self):
                LOGGER.error("Number of %s not matching: %d for %d editions",
                             time, len(times), len(self))
                return False
            if time == 'elapsed_time' and len(times) != len(self)+1:
                LOGGER.error("Number of elasped_time not matching: %d"
                             "(N+1 editions expected, %d editions)",
                             len(times), len(self))
                return False
        return True



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Module to access in easy way results stored in list of dictionaries.

This module is composed of 2 classes:

  * :class:`Browser` that stores the list of dictionaries and builds an
    :class:`Index` to facilitate selections;
  * :class:`Index` based on :class:`collections.defaultdict` to perform
    selections on the list of dictionaries


The classes :class:`Index` and :class:`Browser` are meant to be general even if
they will be shown and used in a specific case: parsing results from Tripoli-4.


The :class:`Index` class
------------------------

This class is based on an inheritance from :class:`collections.abc.Mapping`
from :mod:`collections`. It implements a ``defaultdict(defaultdict(set))`` from
:class:`collections.defaultdict`.

:class:`set` contains `int` that corresponds to the index of the dictionary in
the list of dictionaries.

:class:`Index` is not supposed to be used standalone, but called from
:class:`Browser`, but this is still possible.


The :class:`Browser` class
--------------------------

This class is analogue to a phonebook: it contains an index and the content,
here stored as a list of dictionaries. It commands the index (building and
selections). Examples are shown below.


.. _browser-example:

Building the browser
^^^^^^^^^^^^^^^^^^^^

Let's consider a bunch of friends going to the restaurant and ordering their
menus. For each of them the waiter has to remember their name, under
``'consumer'``, their choice of menu under ``'menu'``, their drink, what they
precisely order as dish under ``'results'`` and optionally the number
corresponding to their choice of dessert. He will represent these orders as a
list of orders, one order being a dictionary.

>>> from valjean.eponine.browser import Browser
>>> from pprint import pprint
>>> orders = [
... {'menu': '1', 'consumer': 'Terry', 'drink': 'beer',
...  'results': {'ingredients_res': ['egg', 'bacon']}},
... {'menu': '2', 'consumer': 'John',
...  'results': [{'ingredients_res': ['egg', 'spam']},
...              {'ingredients_res': ['tomato', 'spam', 'bacon']}]},
... {'menu': '1', 'consumer': 'Graham', 'drink': 'coffee',
...  'results': [{'ingredients_res': ['spam', 'egg', 'spam']}]},
... {'menu': '3', 'consumer': 'Eric', 'drink': 'beer',
...  'results': {'ingredients_res': ['sausage'],
...              'side_res': 'baked beans'}},
... {'menu': 'royal', 'consumer': 'Michael', 'drink': 'brandy', 'dessert': 3,
...  'results': {'dish_res': ['lobster thermidor', 'Mornay sauce']}}]
>>> com_br = Browser(orders)
>>> print(com_br)
Browser object -> Number of content items: 5, data key: 'results', \
available metadata keys: ['consumer', 'dessert', 'drink', 'index', 'menu']
               -> Number of globals: 0

Some global variables can be added, as a dictionary, normally common to all the
results sent under the argument ``content``.

>>> global_vars = {'table': 42, 'service_time': 300, 'priority': -1}
>>> com_br = Browser(orders, global_vars=global_vars)
>>> print(com_br)
Browser object -> Number of content items: 5, data key: 'results', \
available metadata keys: ['consumer', 'dessert', 'drink', 'index', 'menu']
               -> Number of globals: 3
>>> pprint(com_br.globals)
{'priority': -1, 'service_time': 300, 'table': 42}


Selection of a given items or of a list of items from content
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Various methods are available to select one order, depending on requirements:

  * get a new Browser:

    >>> sel_br = com_br.filter_by(menu='1', drink='beer')
    >>> pprint(sel_br.content)  # doctest: +NORMALIZE_WHITESPACE
    [{'consumer': 'Terry',  'drink': 'beer', 'index': 0, 'menu': '1', \
'results': {'ingredients_res': ['egg', 'bacon']}}]

    * check if a key is present or not:

    >>> 'drink' in sel_br
    True
    >>> 'dessert' in sel_br
    False
    >>> 'dessert' in com_br
    True

    The ``'dessert'`` key has been removed from the Browser issued from
    the selection while it is still present in the original one.

  * get the available keys (sorted to be able to test them in the doctest, else
    list is enough):

    >>> sorted(sel_br.keys())
    ['consumer', 'drink', 'index', 'menu']
    >>> sorted(com_br.keys())
    ['consumer', 'dessert', 'drink', 'index', 'menu']

  * if the required key doesn't exist a warning is emitted:

    >>> sel_br = com_br.filter_by(quantity=5)
    >>> # prints  WARNING     browser: quantity not a valid key. Possible \
ones are ['consumer', 'dessert', 'drink', 'index', 'menu']
    >>> 'quantity' in com_br
    False

  * if the value corresponding to the key doesn't exist another warning is
    emitted:

    >>> sel_br = com_br.filter_by(drink='wine')
    >>> # prints  WARNING     browser: wine is not a valid drink

  * to know the available values corresponding to the keys (without the
    corresponding items indexes):

    >>> sorted(com_br.available_values('drink'))
    ['beer', 'brandy', 'coffee']

  * if the key doesn't exist an 'empty generator' is emitted:

    >>> sorted(com_br.available_values('quantity'))
    []

  * to directly get the content items corresponding to the selection, use the
    method :func:`Browser.select_by`

    >>> sel_br = com_br.select_by(consumer='Graham')
    >>> type(sel_br)
    <class 'dict'>
    >>> len(sel_br)
    5
    >>> pprint(sel_br)  # doctest: +NORMALIZE_WHITESPACE
    {'consumer': 'Graham', 'drink': 'coffee', 'index': 2, 'menu': '1', \
'results': [{'ingredients_res': ['spam', 'egg', 'spam']}]}

  * this does not work when several items correspond to the selection:

    >>> sel_br = com_br.select_by(drink='beer')
    Traceback (most recent call last):
            [...]
    valjean.eponine.browser.TooManyItemsBrowserError: Several content items \
correspond to your choice, please refine your selection using additional \
keywords

  * if no item corresponds to the selection another exception is thrown:

    >>> sel_br = com_br.select_by(menu='4')
    Traceback (most recent call last):
            [...]
    valjean.eponine.browser.NoItemBrowserError: No item corresponding to the \
selection.

Module API
----------
'''

import logging
from collections import defaultdict
from collections.abc import Mapping, Container


LOGGER = logging.getLogger(__name__)


def _make_defaultdict_set():
    '''The sole purpose of this function is to give a name to the defaultdict
    factory in :meth:`Index.index`. Without a name, the :class:`Index` class
    cannot be serialized by :mod:`pickle`.
    '''
    return defaultdict(set)


class Index(Mapping):
    '''Class to describe index used in Browser.

    The structure of Index is a ``defaultdict(defaultdict(set))``.  This class
    was derived mainly for pretty-printing purposes.

    Quick example of index (menu for 4 persons, identified by numbers, one has
    no drink):

    >>> from valjean.eponine.browser import Index
    >>> myindex = Index()
    >>> myindex.index['drink']['beer'] = {1, 4}
    >>> myindex.index['drink']['wine'] = {2}
    >>> myindex.index['menu']['spam'] = {1, 3}
    >>> myindex.index['menu']['egg'] = {2}
    >>> myindex.index['menu']['bacon'] = {4}
    >>> myindex.dump(sort=True)
    "{'drink': {'beer': {1, 4}, 'wine': {2}}, \
'menu': {'bacon': {4}, 'egg': {2}, 'spam': {1, 3}}}"
    >>> 'drink' in myindex
    True
    >>> 'consumer' in myindex
    False
    >>> len(myindex)
    2
    >>> for k in sorted(myindex):
    ...    print(k, sorted(myindex[k].keys()))
    drink ['beer', 'wine']
    menu ['bacon', 'egg', 'spam']

    The :func:`keep_only` method allows to get a sub-Index from a given set of
    ids (int), removing all keys not involved in the corresponding ids:

    >>> myindex.keep_only({2}).dump(sort=True)
    "{'drink': {'wine': {2}}, 'menu': {'egg': {2}}}"
    >>> menu_clients14 = myindex.keep_only({1, 4})
    >>> sorted(menu_clients14.keys()) == ['drink', 'menu']
    True
    >>> list(menu_clients14['drink'].keys()) == ['beer']
    True
    >>> list(menu_clients14['drink'].values()) == [{1, 4}]
    True
    >>> sorted(menu_clients14['menu'].keys()) == ['bacon', 'spam']
    True
    >>> menu_clients14['menu']['spam'] == {1}
    True
    >>> 3 in menu_clients14['menu']['spam']
    False
    >>> menu_client3 = myindex.keep_only({3})
    >>> list(menu_client3.keys()) == ['menu']
    True
    >>> 'drink' in menu_client3
    False

    The key ``'drink'`` has been removed from the last index as 2 did not
    required it.

    If you print an :class:`Index`, it looks like a standard dictionary (``{
    ... }`` instead of ``defaultdict(...)``) but the keys are not sorted:

    >>> print(myindex)
    {...: {...: {...}...}}
    '''

    def __init__(self):
        self.index = defaultdict(_make_defaultdict_set)

    def __str__(self):
        lstr = ["{"]
        for i, (key, dset) in enumerate(list(self.index.items())):
            lstr.append(f'{key!r}: {{')
            for j, (dkey, ind) in enumerate(list(dset.items())):
                lstr.append(f'{dkey!r}: {ind!r}')
                if j < len(dset) - 1:
                    lstr.append(', ')
            lstr.append('}')
            if i < len(self.index) - 1:
                lstr.append(', ')
        lstr.append('}')
        return ''.join(lstr)

    def __repr__(self):
        return self.index.__repr__()

    def __getitem__(self, item):
        return self.index.__getitem__(item)

    def __len__(self):
        return len(self.index)

    def __iter__(self):
        return iter(self.index)

    def __contains__(self, key):
        return self.index.__contains__(key)

    def keep_only(self, ids):
        '''Get an :class:`Index` containing only the relevant keywords for the
        required set of ids.

        :param set(int) ids: index corresponding to the required elements of
          the list of content items
        :returns: Index only containing the keys involved in the ids
        '''
        assert isinstance(ids, set)
        lind = Index()
        if not ids:
            return lind
        for key in self.index:
            for kwd, kset in self.index[key].items():
                tmpset = kset & ids
                if tmpset:
                    lind[key][kwd] = tmpset
        return lind

    def dump(self, *, sort=False):
        '''Dump the Index.

        If ``sort == False`` (default case), returns :func:`__str__` result,
        else returns sorted Index (alphabetic order for keys).
        '''
        if sort:
            lstr = ["{"]
            for i, (key, dset) in enumerate(sorted(self.index.items())):
                lstr.append(f'{key!r}: {{')
                for j, (dkey, ind) in enumerate(sorted(dset.items(), key=str)):
                    lstr.append(f'{dkey!r}: {ind!r}')
                    if j < len(dset) - 1:
                        lstr.append(', ')
                lstr.append('}')
                if i < len(self.index) - 1:
                    lstr.append(', ')
            lstr.append('}')
            return ''.join(lstr)
        return str(self)


class Browser(Container):
    '''Class to perform selections on results.

    This class is based on four objects:

      * the content, as a list of dictionaries (containing data and metadata)
      * the key corresponding to data in the dictionary (default='results')
      * an index based on content elements allowing easy selections on each
        metadata
      * a dictionary corresponding to global variables (common to all content
        items).

    Initialization parameters:

    :param list(dict) content: list of items containing data and metadata
    :param str data_key: key in the content items corresponding to results or
      data, that should not be used in index (as always present and mandatory)
    :param dict global_vars: global variables (optional, default=None)

    An additional key is added at the Index construction: ``'index'`` in order
    to keep track of the order of the list and being able to do selection on
    it.

    Examples on development / debugging methods:

    Let's use the example detailled above in
    :ref:`module introduction <browser-example>`:

    >>> from valjean.eponine.browser import Browser
    >>> orders = [
    ... {'menu': '1', 'consumer': 'Terry', 'drink': 'beer',
    ...  'results': {'ingredients_res': ['egg', 'bacon']}},
    ... {'menu': '2', 'consumer': 'John',
    ...  'results': [{'ingredients_res': ['egg', 'spam']},
    ...              {'ingredients_res': ['tomato', 'spam', 'bacon']}]},
    ... {'menu': '1', 'consumer': 'Graham', 'drink': 'coffee',
    ...  'results': [{'ingredients_res': ['spam', 'egg', 'spam']}]},
    ... {'menu': '3', 'consumer': 'Eric', 'drink': 'beer',
    ...  'results': {'ingredients_res': ['sausage'],
    ...              'side_res': 'baked beans'}},
    ... {'menu': 'royal', 'consumer': 'Michael', 'drink': 'brandy',
    ...  'dessert': 3,
    ...  'results': {'dish_res': ['lobster thermidor', 'Mornay sauce']}}]
    >>> com_br = Browser(orders)

    * possibility to get the item id directly (internally used method):

      >>> ind = com_br._filter_items_id_by(drink='coffee')
      >>> isinstance(ind, set)
      True
      >>> print(ind)
      {2}

    * possibility to get the index of the content element stripped without
      rebuilding the full Browser:

      >>> ind = com_br._filter_index_by(menu='1')
      >>> isinstance(ind, Index)
      True
      >>> ind.dump(sort=True)  # doctest: +NORMALIZE_WHITESPACE
      "{'consumer': {'Graham': {2}, 'Terry': {0}}, \
'drink': {'beer': {0}, 'coffee': {2}}, 'index': {0: {0}, 2: {2}}, \
'menu': {'1': {0, 2}}}"

      The 'dessert' key has been stripped from the index:

      >>> 'dessert' in ind
      False

    Debug print is available thanks to :func:`__repr__`:

    >>> small_order = [{'dessert': 1, 'drink': 'beer', 'results': ['spam']}]
    >>> so_br = Browser(small_order)
    >>> f"{so_br!r}"
    "<class 'valjean.eponine.browser.Browser'>, (Content items: ..., \
Index: ...)"
    '''

    def __init__(self, content, data_key='results', global_vars=None):
        self.content = [r.copy() for r in content]
        self.data_key = data_key
        self.index = self._build_index()
        LOGGER.debug("Index: %s", self.index)
        self.globals = (global_vars.copy() if isinstance(global_vars, dict)
                        else {})

    def __eq__(self, other):
        return (self.content == other.content
                and self.data_key == other.data_key
                and self.globals == other.globals)

    def __ne__(self, other):
        return not self == other

    def _build_index(self):
        '''Build index from all content elements in the list.

        Keys of the sets are keywords used to describe the items and/or the
        scores (if flat case).

        :param str data_key: key in list of content items corresponding to
          results or data
        :returns: :class:`Index`
        '''
        index = Index()
        for ielt, elt in enumerate(self.content):
            elt['index'] = ielt
            for key in elt:
                if key != self.data_key:
                    index[key][elt[key]].add(ielt)
        return index

    def __contains__(self, key):
        if key in self.index:
            return True
        return False

    def __len__(self):
        return len(self.content)

    def is_empty(self):
        '''Check if the Browser is empty or not.

        Empty meaning no elements in content AND no globals.
        '''
        return not self.content and not self.globals

    def merge(self, other):
        '''Merge two browsers.

        This method merge 2 browsers: the *other* one appears then at the
        end of the *self* one. Global variables are also merged. The new index
        correspond to the merged case.

        :param Browser other: another browser
        :rtype: Browser
        '''
        if self.data_key != other.data_key:
            raise ValueError('Same data_key is required to merge Browsers')
        if self.globals != other.globals:
            LOGGER.debug('globals will be updated with other values')
        new_glob = self.globals.copy()
        new_glob.update(other.globals)
        new_content = self.content + other.content
        LOGGER.debug("Nb self items (%d) + Nb other items (%d) = %d",
                     len(self), len(other), len(new_content))
        return Browser(new_content, data_key=self.data_key,
                       global_vars=new_glob)

    def keys(self):
        '''Get the available keys in the index (so in the items list). As
        usual it returns a generator.
        '''
        return tuple(self.index.keys())

    def available_values(self, key):
        '''Get the available keys in the *second* level, i.e. under the given
        external one.

        :param str key: 'external' key (from outer defaultdict)
        :returns: generator with corresponding keys (or empty one)
        '''
        if key in self:
            return tuple(self.index[key].keys())
        return ()

    def __str__(self):
        cls_name = self.__class__.__name__
        return (f"{cls_name} object -> "
                f"Number of content items: {len(self.content)}, "
                f"data key: {self.data_key!r}, "
                f"available metadata keys: {sorted(self.keys())}\n"
                f"{'':>{len(cls_name)}}        "
                f"-> Number of globals: {len(self.globals)}")

    def __repr__(self):
        return (f"{self.__class__}, (Content items: {self.content!r}, "
                f"Index: {self.index!r})")

    def _filter_items_id_by(self, **kwargs):
        '''Selection of content items indices according to kwargs criteria.

        :param \\**\\kwargs: keyword arguments to specify the required
          response. More than one are allowed.
        :return: set of ids
        :rtype: set(int)
        '''
        itemids = set(range(len(self.content)))
        for kwd, kwarg in kwargs.items():
            if kwd not in self.index:
                LOGGER.warning("%s not a valid key. Possible ones are %s",
                               kwd, sorted(self.keys()))
                return set()
            if kwarg not in self.index[kwd]:
                LOGGER.warning("%s is not a valid %s", kwarg, kwd)
                return set()
            itemids = itemids & self.index[kwd][kwarg]
        if not itemids:
            LOGGER.warning("Wrong selection, item might be not present. "
                           "Also check if requirements are consistent.")
            return set()
        return itemids

    def _filter_index_by(self, **kwargs):
        '''Get index corresponding to selection given thanks to keyword
        arguments.

        :param \\**\\kwargs: keyword arguments to specify the required item.
            More than one are allowed.
        :returns: :class:`Index` (stripped from useless keys)
        '''
        itemids = self._filter_items_id_by(**kwargs)
        return self.index.keep_only(itemids)

    def filter_by(self, include=(), exclude=(), **kwargs):
        '''Get a Browser corresponding to selection from keyword
        arguments.

        :param \\**\\kwargs: keyword arguments to specify the required item.
            More than one are allowed.
        :param tuple(str) include: metadata keys required in the content items
            but for which the value is not necessarly known
        :param tuple(str) exclude: metadata that should not be present in the
            items and for which the value is not necessarly known
        :returns: :class:`Browser` (subset of the default one, corresponding to
            the selection)
        '''
        LOGGER.debug("in select_by, kwargs=%s", kwargs)
        sincl, sexcl = set(include), set(exclude)
        respids = self._filter_items_id_by(**kwargs)
        lresp = [self.content[i] for i in sorted(respids)
                 if sincl.issubset(self.content[i])
                 and not sexcl.intersection(self.content[i])]
        sub_br = Browser(lresp, global_vars=self.globals)
        return sub_br

    def select_by(self, *, include=(), exclude=(), **kwargs):
        '''Get an item or the list of items from content corresponding to
        selection from keyword arguments.

        :param \\**\\kwargs: keyword arguments to specify the required items.
            More than one are allowed.
        :param tuple(str) include: metadata keys required in the items but for
            which the value is not necessarly known
        :param tuple(str) exclude: metadata that should not be present in the
          items and for which the value is not necessarly known
        :raises NoItemBrowserError: if no item corresponds to the selection
        :raises TooManyItemsBrowserError: if more than one item corresponds to
          the provided keywords
        :rtype: dict
        '''
        respids = self._filter_items_id_by(**kwargs)
        sincl, sexcl = set(include), set(exclude)
        litems = [self.content[i] for i in sorted(respids)
                  if sincl.issubset(self.content[i])
                  and not sexcl.intersection(self.content[i])]
        if not litems:
            raise NoItemBrowserError("No item corresponding to the selection.")
        if len(litems) > 1:
            raise TooManyItemsBrowserError(
                "Several content items correspond to your choice, "
                "please refine your selection using additional keywords")
        return litems[0]


class TooManyItemsBrowserError(LookupError):
    '''Error to :class:`Browser` when too many items correspond to the
    requested selection.
    '''


class NoItemBrowserError(LookupError):
    '''Error to :class:`Browser` when no item corresponds to the selection.'''



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''A reversible list (:class:`RList` for short) keeps track of the indices of
its elements for fast reverse lookup. It has mostly the same semantics as
lists:

    >>> from valjean.cosette.rlist import RList
    >>> dead, stiff, bereft, rests = ('dead', 'stiff', 'bereft of life',
    ...                               'rests in peace')
    >>> parrot = RList([dead, stiff, bereft, rests])
    >>> parrot[0]
    'dead'
    >>> parrot[-2]
    'bereft of life'
    >>> del parrot[1]
    >>> print(parrot)
    ['dead', 'bereft of life', 'rests in peace']
    >>> parrot == ['dead', 'bereft of life', 'rests in peace']
    True
    >>> parrot == RList(['a stiff'])
    False

Additionally, you can quickly (in `O(1)` time on average) look up the index of
an item or check if an item is contained in the list:

    >>> parrot.index(rests)  # this call is O(1)
    2
    >>> rests in parrot  # and so is this
    True

This operation takes `O(n)` average time in normal lists.

The most important differences with respect to the semantics of traditional
Python lists are:

  1. Slicing operations are not supported.
  2. The notion of "containment" is user-defined. By default, a value belongs
     to an :class:`RList` if it has the same `key` as one of the list elements.
     The `key`, by default, is the value :func:`id`.  In other words, normal
     lists compare items with the ``==`` operator, while :class:`RList`
     compares object IDs by default.  Objects that compare equal (with the
     standard ``==`` operator) may be used interchangeably in a list, but not
     in a :class:`RList`.  For example, here is a simple class:

        >>> class A:
        ...    def __init__(self, x):
        ...        self.x = x
        ...    def __eq__(self, other):
        ...        return self.x == other.x

     and here is how it behaves in a normal list:

        >>> a1 = A(42)
        >>> a2 = A(42)
        >>> a1 == a2
        True
        >>> a1 in [a2]
        True

     The objects ``a1`` and ``a2`` compare equal, so to the eyes of the list
     ``[a2]`` they are "the same". :class:`RList`, on the other hand, does not
     play along with this charade by default:

        >>> a1 is a2
        False
        >>> id(a1) == id(a2)  # equivalent to the line above
        False
        >>> a1 in RList([a2])
        False

     Since ``a1`` and ``a2`` are distinct objects (they live in different
     memory locations), they are different in the eyes of :class:`RList`.  This
     may result in unexpected behaviour, especially with strings, ints or other
     small objects for which the Python interepreter may provide some kind of
     caching optimisation:

        >>> lst = RList([1])
        >>> 1 in lst
        True
        >>> lst = RList([1234567])
        >>> 1234567 in lst
        False
        >>> # wat?
        >>> 1234567 in RList([1234567])
        True
        >>> # WAT?

     This weird behaviour is actually a well-documented quirk of the CPython
     implementation.

     If you want, you can define your own notion of `key` for your objects by
     passing a suitable function to the ``key`` argument of the :class:`RList`
     constructor; the only constraint is that the value returned by `key` must
     be hashable. For example, if you have a list of strings, you can use the
     string itself as a key as follows:

        >>> parrot_by_value = RList(['Norwegian blue', 'plumage', 'pining'],
        ...                         key=lambda x: x)
        >>> 'plumage' in parrot_by_value
        True
        >>> 'bereft of life' in parrot_by_value
        False

     You can also use more sophisticated `key` functions:

        >>> a_rlist = RList([A(0), A(1), A(2)], key=lambda a: a.x)
        >>> A(2) in a_rlist
        True
        >>> A(5) in a_rlist
        False
'''

from collections import defaultdict
from collections.abc import MutableSequence


class RList(MutableSequence):
    '''Create a reversible list from an iterable.

    :param args: An iterable containing some elements.
    '''

    def __init__(self, args=None, *, key=id):
        self._seq = []
        self._index = defaultdict(list)
        self._key = key
        if args is not None:
            for element in args:
                self.append(element)

    def __getitem__(self, index):
        return self._seq[index]

    def __setitem__(self, index, value):
        if index < 0:
            index += len(self)
        old_id = self._key(self._seq[index])
        indices = self._index[old_id]
        indices.remove(index)
        if not indices:
            del self._index[old_id]
        self._seq[index] = value
        self._index[self._key(value)].append(index)

    def __delitem__(self, index):
        if index < 0:
            index += len(self)
        # Decrease all mapped indices
        delete_ids = []
        for obj_id, indices in self._index.items():
            filtered = filter(lambda i: i != index, indices)
            new_indices = [i if i < index else i-1 for i in filtered]
            if new_indices:
                self._index[obj_id] = new_indices
            else:
                # prune empty index lists
                delete_ids.append(obj_id)
        for obj_id in delete_ids:
            del self._index[obj_id]
        del self._seq[index]

    def __len__(self):
        return len(self._seq)

    def __repr__(self):
        return f'RList({self._seq!r})'

    def __str__(self):
        return self._seq.__str__()

    def __eq__(self, other):
        if isinstance(other, list):
            return self._seq == other
        if isinstance(other, RList):
            return self._seq == other._seq
        raise TypeError('RList can only be compared to lists or RLists')

    def __ne__(self, other):
        return not self == other

    def insert(self, index, value):
        '''Insert an element at a given index.'''
        n_elems = len(self)
        if index < 0:
            index += n_elems
        # Normalize the index so that it is valid in the extended list.  Things
        # like the following
        #   >>> lst = [1,2,3]
        #   >>> lst.insert(42, 4)
        # are allowed with Python lists, so we want to allow them with RLists,
        # too.
        index = min(n_elems, max(0, index))
        # Increase all mapped indices
        for obj_id, indices in self._index.items():
            new_indices = [i if i < index else i+1 for i in indices]
            self._index[obj_id] = new_indices
        self._seq.insert(index, value)
        self._index[self._key(value)].append(index)

    def index(self, value, start=0, stop=None):
        '''Return the index of the given value, if present.

        :param value: The object to search for.
        :param int start: The index to search from.
        :param int stop: The index to search up to.
        :raises ValueError: if the element is not present in the container.
        :returns: The index of (the first occurrence of) `value` in the list.
                  The returned value `i` always satisfies `start <= i < stop`.
        '''
        indices = self._index.get(self._key(value), None)
        if indices is None:
            raise ValueError(f'{value} is not in list')
        found = next((ind for ind in indices
                      if ind >= start and (stop is None or ind < stop)),
                     None)
        if found is None:
            raise ValueError(f'{value} is not in list')
        return found

    def indices(self, value):
        '''Return all the indices of the given value, if present.

        :param value: The object to search for.
        :raises KeyError: if the element is not present in the container.
        :returns: All the list indices where `value` occurs.
        '''
        i = self._index.get(self._key(value), None)
        if i is None:
            raise KeyError(self._key(value))
        return i

    def get_index(self, value, default):
        '''Return the index of the given value, or a default if the value is
        missing.

        :param value: The object to search for.
        :param default: The value to be returned if `value` is not present in
                        the container.
        '''
        ind = self._index.get(self._key(value), None)
        if ind is None:
            return default
        return ind[0]

    def __contains__(self, value):
        return self._key(value) in self._index

    def swap(self, i, j):
        '''Swap two elements of the list.

        After this operation, the ith and jth elements will be swapped.

        :param int i: The index of the first element.
        :param int j: The index of the second element.
        '''
        tmp = self[i]
        self[i] = self[j]
        self[j] = tmp

    def copy(self):
        '''Return a shallow copy of this object.'''
        return RList(self._seq)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module contains the :class:`Formatter` abstract base class. All other
formatters must derive from it.
'''

from abc import ABC, abstractmethod


class Formatter(ABC):
    '''Abstract base class for any formatter.'''

    @abstractmethod
    def header(self, name, depth):
        '''Format a section header.'''
        raise NotImplementedError('header() method is not implemented')

    def text(self, text):
        '''Format some text.'''
        raise NotImplementedError('text() method is not implemented')

    def template(self, item):
        '''Convert an item to the relevant format.'''
        class_name = item.__class__.__name__
        meth_name = 'format_' + class_name.lower()
        try:
            meth = getattr(self, meth_name)
        except AttributeError:
            return ''
        return meth(item)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module is used to dump the results of the parsing in a pretty way. It
is mainly meant for debugging.

Each function returns a string that is printed in the LOGGER in the main
function, :func:`print_result` that acts directly on the
``pyparsing.ParseResults``. This main function is called in a
``pyparsing.ParseElement.setParseAction`` in :mod:`~grammar`.
'''

import logging
from collections import OrderedDict
import numpy as np


LOGGER = logging.getLogger(__name__)
MAX_DEPTH = 0
MAX_DEPTH_STR = f"MAX_DEPTH = {MAX_DEPTH} reached\n"


def array_to_str(array):
    '''Transform :obj:`numpy.ndarray` in condensed string.

    :param numpy.ndarray array: array to print
    :returns: str
    '''
    lstr = []
    if array.shape != ():
        lstr.append(f"{type(array)}, shape: {array.shape}, "
                    f"dtype: {array.dtype}, squeezed:")
        lstr.append(np.array2string(np.squeeze(array), precision=6,
                                    suppress_small=True,
                                    formatter={'float_kind': '{:.6e}'.format}))
    else:
        arr_str = np.array2string(array, precision=6, suppress_small=True,
                                  formatter={'float_kind': '{:.6e}'.format})
        lstr.append(f"{arr_str}, dtype: {array.dtype}")
    return '\n'.join(lstr)


def result_to_str_according_type(res, depth=0):
    '''Choose function to be used for printing according to ``res`` type.

    :param res: interpreted result from `pyparsing`
    :type res: dict, list, numpy.ndarray
    :param int depth: level of prints
    :const MAX_DEPTH: maximum of prints level
    '''
    if depth > MAX_DEPTH:
        return MAX_DEPTH_STR
    spaces = "  "*depth
    lstr = []
    if isinstance(res, dict):
        lstr.append(dict_to_str(res, depth))
    elif isinstance(res, list):
        lstr.append(list_to_str(res, depth))
    elif isinstance(res, (np.ndarray, np.generic)):
        lstr.append(array_to_str(res))
    else:
        lstr.append(f"{spaces}{res!s}")
    return '\n'.join(lstr)


def dict_to_str(diction, depth=0):
    '''Convert dictionary to customised string.

    :param dict diction: python dictionary
    :param int depth: level of prints
    :const MAX_DEPTH: maximum of prints level
    :returns: str
    '''
    spaces = "  "*depth
    lstr = []
    dictkeys = (list(diction.keys()) if isinstance(diction, OrderedDict)
                else sorted(diction))
    lstr.append(f"{spaces}Dict with keys = {dictkeys}")
    if depth > MAX_DEPTH:
        return MAX_DEPTH_STR
    for key in dictkeys:
        depth += 1
        spaces = "  "*depth
        key_str = spaces + key
        if isinstance(diction[key], (dict, list, np.ndarray, np.generic)):
            res_str = result_to_str_according_type(diction[key], depth)
            lstr.append(f"{key_str} {res_str}")
        else:
            lstr.append(f"{key_str} {diction[key]}")
        depth -= 1
    return '\n'.join(lstr)


def list_to_str(liste, depth=0):
    '''Convert list to customised string.

    :param list liste: python list to be printed
    :param int depth: level of prints
    :const MAX_DEPTH: maximum of prints level
    :returns: str
    '''
    spaces = "  "*depth
    lstr = []
    lstr.append(f"{spaces}list of {len(liste)} elements -> ")
    if isinstance(liste[0], (dict, list, np.ndarray)):
        if depth > MAX_DEPTH:
            lstr.append(MAX_DEPTH_STR)
            return '\n'.join(lstr)
        for ielt, elt in enumerate(liste):
            depth += 1
            res_str = result_to_str_according_type(elt, depth)
            lstr.append(f"{spaces}elt {ielt}: {res_str}")
            depth -= 1
    else:
        lstr.append(f"{spaces}{liste}")
    return '\n'.join(lstr)


def response_to_str(res, depth=0):
    '''Convert response to customised string.

    Rigid structure as it is supposed to contain only 2 kinds of "data":
    the results or really data under the key ``'results'`` and the associated
    metadata under all the other keys (can be numerous). ``res`` is thus a
    dictionary. This is made explicit in the printing code.

    :param dict res: response part of the output dictionary
    :param int depth: level of prints
    :const MAX_DEPTH: maximum of prints level
    :returns: str
    '''
    if depth > MAX_DEPTH:
        return MAX_DEPTH_STR
    assert 'results' in res.keys()
    spaces = "  "*depth
    lstr = []
    # First print the metadata (in alphabetic order)
    lstr.append(spaces + "'response metadata'")
    lstr.append(dict_to_str({k: v for k, v in res.items() if k != 'results'},
                            depth))
    # Then print the results
    lstr.append(f"{spaces}results ({type(res['results'])}) -> "
                f"{res['response_type']}")
    rres = res['results']
    depth += 1
    if not isinstance(rres, (list, dict)):
        if isinstance(rres, np.ndarray):
            lstr.append(array_to_str(rres))
        else:
            lstr.append(str(rres))
    elif isinstance(rres, list):
        lstr.append(list_to_str(rres, depth))
    else:
        lstr.append(dict_to_str(rres, depth))
    depth -= 1
    return '\n'.join(lstr)


def parsing_result_to_str(toks):
    '''Convert parsing result in customised string.

    :param toks: `pyparsing` result
    :type toks: |parseres|
    :const MAX_DEPTH: maximum of prints level
    '''
    depth = 0
    lstr = []
    intro_str = "\n" + "-"*30 + " Structured parsed result " + "-"*30
    lstr.append(intro_str)
    for res in toks:
        depth += 1
        if depth > MAX_DEPTH:
            break
        lstr.append(f"\nKeys: {sorted(res.keys())}")
        for key in sorted(res):
            depth += 1
            if depth > MAX_DEPTH:
                break
            lstr.append(f"\n{key} ")
            if key == 'list_responses':
                lstr.append(f"\nNumber of responses: {len(res[key])}")
                for iresp, resp in enumerate(res[key]):
                    depth += 1
                    spaces = "  "*depth
                    lstr.append(f"\n\n{spaces}RESPONSE {iresp}")
                    lstr.append(response_to_str(resp, depth))
                    depth -= 1
            else:
                lstr.append(result_to_str_according_type(res[key], depth))
            depth -= 1
            lstr.append('\n')
        depth -= 1
    lstr.append("-"*80)
    return ''.join(lstr)


def dump_in_logger(toks):
    ''''Dump the parsing result thanks to LOGGER.

    Print will be done only in DEBUG mode (logger), if MAX_DEPTH != 0.

    .. todo:: link to logger documentation
    '''
    if not LOGGER.isEnabledFor(logging.DEBUG):
        return
    LOGGER.debug("Number of results: %d", len(toks))
    LOGGER.debug(parsing_result_to_str(toks))


def dump_using_print(toks):
    '''Dump the parsing result thanks to print.'''
    print(parsing_result_to_str(toks))



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

r'''
.. _family-wise: https://en.wikipedia.org/wiki/Family-wise_error_rate
.. _Bonferroni: https://en.wikipedia.org/wiki/Bonferroni_correction
.. _Holm-Bonferroni: https://en.wikipedia.org/wiki/Holm-Bonferroni_method

In some cases we would like to interpret not only one test (like on a generic
score in Tripoli-4) but multiple ones (in a spectrum case for example). Each
individual test is supposed to be successful but in most of the case we can
accept that some of them fail (in a given acceptance of course).

The Bonferroni correction and the Holm-Bonferroni method allow to treat such
cases. See for example `family-wise`_ error rate Wikipedia webpage for more
details on this kind    of tests.

Here are implemented two tests based on comparison of the p-value from the
chosen test and a significance level:

* Bonferroni correction: all individual p-values are compared to the same
  significance level, depending on the number of tests;
* Holm-Bonferroni method: p-values are first sorted then compared to different
  significance levels depending on number of tests and rank.


In both cases a first test is chosen, like Student test, to evaluate the
individual hypothesis. The p-value of the test is calculated and used in the
following tests. Like in other tests, we consider two-sided significance levels
here.

The null hypothesis is that the p-value is higher than a given significance
level, that depends on the overall required significance level, the number of
tests considered and the chosen method (Bonferroni or Holm-Bonferroni).


Bonferroni correction
`````````````````````

The Bonferroni_ correction is used for multiple comparisons. It is based on a
weighted comparison to the significance level α.

If we have :math:`m` different tests (for example a spectrum with :math:`m`
bins), the significance level α will be weighted by :math:`m`. This p-value of
the :math:`m` tests will then be compared to that value. As a consequence, the
null hypothesis will be rejected if

.. math::

    p_k \leq \frac{\alpha}{m}

The  individual significance level is usually largely smaller than the required
one and the sum of them obvioulsy give α.

Conclusion of the test depends on what we want:

* it can be a simple ``True`` if all hypothesis are accepted;
* it can be ``False`` if at least one null hypothesis was rejected;
* we can also choose a level of acceptance, for example "it is fine if 1 % of
  the null hypothesis are rejected"

Per default, in our case the test is rejected if there is at least one null
hypothesis rejected.


Holm-Bonferroni method
``````````````````````

The Holm-Bonferroni_ method is a variation of the Bonferroni correction.

The p-value from the chosen test are first sorted from lower to higher value.
Then each of them are compared to the significance level weighted by
:math:`m - k + 1` with :math:`m` the number of tests and :math:`k` the rank
(starting at 1). The null hypothesis is rejected if

.. math::

    p_k \leq \frac{\alpha}{m - k + 1}.

The test can be stopped at the first :math:`k` accepting the null hypothesis.

This test is quite conservative. It can be possible to get the proportion of
accepted and rejected hypothesis, like in the Bonferroni correction case.


Code implementation
```````````````````

In both cases, the test is initiliazed with another test, a Student test for
example. This test can be evaluated inside the new test
:meth:`~TestBonferroni.evaluate` method or outside of it. Only examples with
the full structure will be shown here.

TestResults are returned as ``False`` if one null hypothesis is rejected.
Access to the initial test is still provided.


Examples
````````

Let's do the usual imports then use the quick example from Student test, with
the p-value calculation:

>>> import numpy as np
>>> from valjean.eponine.dataset import Dataset
>>> from valjean.gavroche.stat_tests.student import TestStudent
>>> from valjean.gavroche.stat_tests.bonferroni import TestBonferroni
>>> from valjean.gavroche.stat_tests.bonferroni import TestHolmBonferroni


Successful test with Bonferroni correction and Holm-Bonferroni method
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

>>> ds1 = Dataset(np.array([5.2, 5.3, 5.25, 5.4, 5.5]),
...               np.array([0.2, 0.25, 0.1, 0.2, 0.3]))
>>> ds2 = Dataset(np.array([5.1, 5.6, 5.2, 5.3, 5.2]),
...               np.array([0.1, 0.3, 0.05, 0.4, 0.3]))
>>> tstudent_12 = TestStudent(ds1, ds2, name="comp",
...                           description="Comparison using Student test",
...                           alpha=0.05, ndf=1000)

>>> tbonf = TestBonferroni(name="bonf", description="Bonferroni correction",
...                        test=tstudent_12, alpha=0.05)
>>> tbonf.ntests
5
>>> tbonf.ntests == ds1.size
True
>>> tbonf.bonf_signi_level
0.005
>>> tbonf_res = tbonf.evaluate()
>>> bool(tbonf_res)
True
>>> tbonf_res.nb_rejected
[0]

>>> tholmbonf = TestHolmBonferroni(name="holm-bonf",
...                                description="Holm-Bonferroni method",
...                                test=tstudent_12, alpha=0.05)
>>> tholmbonf.alpha  # alpha is two-sided
0.025
>>> thb_res = tholmbonf.evaluate()
>>> bool(thb_res)
True
>>> thb_res.nb_rejected
[0]


Failing test with Bonferroni correction and Holm-Bonferroni method
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

>>> ds3 = Dataset(np.array([5.1, 5.9, 5.8, 5.3, 4.6]),
...               np.array([0.1, 0.1, 0.05, 0.4, 0.12]))
>>> tstudent_13 = TestStudent(ds1, ds3, name="comp",
...                           description="Comparison using Student test",
...                           alpha=0.05, ndf=1000)

>>> tbonf = TestBonferroni(name="bonf", description="Bonferroni correction",
...                        test=tstudent_13, alpha=0.05)
>>> tbonf.ntests
5
>>> tbonf.bonf_signi_level
0.005
>>> tbonf_res = tbonf.evaluate()
>>> bool(tbonf_res)
False
>>> tbonf_res.nb_rejected
[1]
>>> tbonf_res.rejected_proportion  # in percentage
[20.0]

>>> tholmbonf = TestHolmBonferroni(name="holm-bonf",
...                                description="Holm-Bonferroni method",
...                                test=tstudent_13, alpha=0.05)
>>> tholmbonf.alpha  # alpha is two-sided
0.025
>>> thb_res = tholmbonf.evaluate()
>>> bool(thb_res)
False
>>> thb_res.nb_rejected
[2]
>>> thb_res.rejected_proportion  # in percentage
[40.0]

For an easier comparison, p-values, associated significance levels and the
corresponding result from the test (rejection or not) can be printed. The
sorting order is also given for the Holm-Bonferroni method.

>>> it = np.nditer([thb_res.first_test_res.pvalue,
...                 tbonf.bonf_signi_level, tbonf_res.rejected_null_hyp,
...                 thb_res.sort_ordering, thb_res.alphas_i,
...                 thb_res.rejected_null_hyp],
...                flags=['multi_index'])
>>> while not it.finished:
...     if it.iterindex == 0:
...         print("index  | pvalue(B) |   α(B)    |  reject B | order(HB) |"
...               f"   α(HB)   | reject HB\n{'-':->77}")
...     print(f"{it.multi_index} | {it[0]:.3e} | {it[1]:.3e} | {it[2]!s:^9} | "
...           f"{it[3]:^7}   | {it[4]:.3e} |   {it[5]!s:<}")
...     _ = it.iternext()
index  | pvalue(B) |   α(B)    |  reject B | order(HB) |   α(HB)   | reject HB
-----------------------------------------------------------------------------
(0, 0) | 6.548e-01 | 5.000e-03 |   False   |    3      | 1.250e-02 |   False
(0, 1) | 2.608e-02 | 5.000e-03 |   False   |    2      | 8.333e-03 |   False
(0, 2) | 1.015e-06 | 5.000e-03 |   True    |    0      | 5.000e-03 |   True
(0, 3) | 8.231e-01 | 5.000e-03 |   False   |    4      | 2.500e-02 |   False
(0, 4) | 5.447e-03 | 5.000e-03 |   False   |    1      | 6.250e-03 |   True


The first index corresponds to the index of the compared dataset in the
``datasets`` container in the input test (compared to ``dsref``).

The Holm-Bonferroni test is uniformly more powerful than the Bonferroni test.
In this particular case, Holm-Bonferroni is able to reject an additional null
hypothesis with respect to plain Bonferroni.


Tests with multi-dimension datasets
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

>>> ds4 = Dataset(np.array([[5.2, 5.3, 5.25], [5.4, 5.5, 5.2]]),
...               np.array([[0.2, 0.25, 0.1], [0.2, 0.3, 0.25]]))
>>> ds5 = Dataset(np.array([[5.1, 5.9, 5.8], [5.3, 4.7, 5.5]]),
...               np.array([[0.1, 0.1, 0.05], [0.4, 0.05, 0.2]]))
>>> ds4.shape
(2, 3)
>>> ds4.size
6
>>> tstudent_45 = TestStudent(ds4, ds5, name="comp",
...                           description="Comparison using Student test",
...                           alpha=0.05, ndf=1000)

>>> tbonf = TestBonferroni(name="bonf", description="Bonferroni correction",
...                        test=tstudent_45, alpha=0.05)
>>> tbonf.ntests
6
>>> print(f'{tbonf.bonf_signi_level:.6f}')
0.004167
>>> tbonf_res = tbonf.evaluate()
>>> bool(tbonf_res)
False
>>> tbonf_res.nb_rejected
[1]
>>> print(f'{tbonf_res.rejected_proportion[0]:.3f} %')
16.667 %

>>> tholmbonf = TestHolmBonferroni(name="holm-bonf",
...                                description="Holm-Bonferroni method",
...                                test=tstudent_45, alpha=0.05)
>>> thb_res = tholmbonf.evaluate()
>>> tholmbonf.alpha
0.025
>>> bool(thb_res)
False
>>> thb_res.nb_rejected
[1]
>>> print(f'{thb_res.rejected_proportion[0]:.3f} %')
16.667 %

The detailed comparison also works for multi-dimensions arrays:

>>> it = np.nditer([thb_res.first_test_res.pvalue, tbonf.bonf_signi_level,
...                 tbonf_res.rejected_null_hyp, thb_res.sort_ordering,
...                 thb_res.alphas_i, thb_res.rejected_null_hyp],
...                flags=['multi_index'])
>>> while not it.finished:
...     if it.iterindex == 0:
...         print("index     | pvalue(B) |   α(B)   | reject B | order(HB) |"
...               f"   α(HB)  | reject HB\n{'-':->79}")
...     print(f"{it.multi_index} |  {it[0]:.2e} | {it[1]:.2e} | {it[2]!s:^7}  "
...           f"| {it[3]:^7}   | {it[4]:.2e} |   {it[5]!s:<}")
...     _ = it.iternext()
index     | pvalue(B) |   α(B)   | reject B | order(HB) |   α(HB)  | reject HB
-------------------------------------------------------------------------------
(0, 0, 0) |  6.55e-01 | 4.17e-03 |  False   |    4      | 1.25e-02 |   False
(0, 0, 1) |  2.61e-02 | 4.17e-03 |  False   |    2      | 6.25e-03 |   False
(0, 0, 2) |  1.01e-06 | 4.17e-03 |  True    |    0      | 4.17e-03 |   True
(0, 1, 0) |  8.23e-01 | 4.17e-03 |  False   |    5      | 2.50e-02 |   False
(0, 1, 1) |  8.66e-03 | 4.17e-03 |  False   |    1      | 5.00e-03 |   False
(0, 1, 2) |  3.49e-01 | 4.17e-03 |  False   |    3      | 8.33e-03 |   False

The first index corresponds to the index of the compared dataset in the
``datasets`` container in the input test (compared to ``dsref``). The next ones
correspond to the shape of the arrays.

The sorting order for Holm-Bonferroni considers all elements as equivalent (not
done per dimension).
'''

import numpy as np
from ..test import Test, TestResult


class TestResultBonferroni(TestResult):
    '''Test result from Bonferroni correction.

    The test returns ``True`` if the test is successful (null hypothesis was
    accepted) and ``False`` if it is rejected.
    '''

    def __init__(self, test, first_test_res, rejected_null_hyp):
        '''Initialisation of :class:`~.TestResultBonferroni`

        :param test: the used Bonferroni test
        :type test: :class:`~.TestBonferroni`
        :param first_test_res: the test result used to obtain the p-values
        :type first_test_res: :class:`~valjean.gavroche.test.TestResult`
                              child class
        :param reject_hyp: rejection of the null hypothesis for all bins
        :type reject_hyp: :class:`numpy.ndarray` (:obj:`bool`)
        '''
        super().__init__(test)
        self.first_test_res = first_test_res
        self.rejected_null_hyp = rejected_null_hyp

    def oracles(self):
        '''Final test (on list of tests, espacially in case of mulitple tests
        on common reference).

        :returns: list(bool)
        '''
        return [not np.any(rnh) for rnh in self.rejected_null_hyp]

    def __bool__(self):
        return not np.any(self.rejected_null_hyp)

    @property
    def nb_rejected(self):
        '''Number of rejected null hypothesis according Holm-Bonferroni test.

        :returns: list(int) or
            :class:`list` (:obj:`numpy.generic` (:obj:`int`))
        '''
        return [np.count_nonzero(rnh) for rnh in self.rejected_null_hyp]

    @property
    def rejected_proportion(self):
        '''Rejected proportion in percentage.

        :returns: :class:`list` (:class:`numpy.generic` (:obj:`float`))
        '''
        return [nr / rnh.size * 100
                for nr, rnh in zip(self.nb_rejected, self.rejected_null_hyp)]


class TestBonferroni(Test):
    '''Bonferroni correction for multiple tests of the same hypothesis.'''

    def __init__(self, *, name, description='', labels=None, test, alpha=0.01):
        '''Initialisation of :class:`TestBonferroni`.


        :param str name: local name of the test
        :param str description: specific description of the test
        :param dict labels: labels to be used for test classification in
                            reports (for example category, input file name,
                            type of result, ...)
        :param test: test from which pvalues will be extracted
        :type test: :class:`valjean.gavroche.test.Test` child class
        :param float alpha: required significance level

        The significance level is considered two-sided here, so divide by 2.
        It is the significance level for the whole test.
        '''
        super().__init__(name=name, description=description, labels=labels)
        self.test = test
        self.alpha = np.float_(alpha / 2)
        self.bonf_signi_level = self.alpha / self.ntests

    @property
    def ntests(self):
        '''Returns the number of hypotheses to test, i.e. number of bins here.
        '''
        return self.test.dsref.size

    @staticmethod
    def bonferroni_correction(pvalues, bonf_signi_level):
        '''Bonferroni correction.

        :param pvalues: p-values from the original test
        :type pvalues: :obj:`numpy.ndarray` (:obj:`float`)
        :param float bonf_signi_level: significance level for each test (α
                                       weighted by the number of tests)
        :returns: :obj:`numpy.ndarray` (:obj:`bool`)

        Compares the p-value to the "Bonferroni significance level":

        * if lower, null hypothesis is rejected for the bin
        * if higher, null hypothesis is accepted for the bin.
        '''
        reject = pvalues <= bonf_signi_level
        return reject

    def evaluate(self):
        '''Evaluate the Bonferroni correction.

        :returns: :class:`~TestResultBonferroni`
        '''
        test_res = self.test.evaluate()
        null_hyp_reject = [
            self.bonferroni_correction(pval, self.bonf_signi_level)
            for pval in test_res.pvalue]
        return TestResultBonferroni(self, test_res, null_hyp_reject)

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        yield from self.test.data()
        yield float(self.alpha).hex().encode('utf-8')


class TestResultHolmBonferroni(TestResult):
    '''Result from the Holm-Bonferroni method.'''

    def __init__(self, test, first_test_res, alphas_i, rejected_null_hyp):
        '''Initialisation of :class:`~.TestResultHolmBonferroni`

        :param test: the used Holm-Bonferroni test
        :type test: :class:`~TestHolmBonferroni`
        :param first_test_res: the test result used to obtain the p-values
        :type first_test_res: :class:`~valjean.gavroche.test.TestResult`
                              child class
        :param sorted_indices: indices of the p-values sorted to get p-values
                               in increasing order
        :type sorted_indices: :obj:`numpy.generic` (:obj:`int`)
        :param alphas_i: significance levels for sorted pvalues
        :type alphas_i: :obj:`numpy.ndarray` (:obj:`float`)
        :param rejected_hyp: rejection of the null hypothesis for all bins
        :type rejected_hyp: :obj:`numpy.ndarray` (:obj:`bool`)
        '''
        super().__init__(test)
        self.alphas_i = alphas_i
        self.rejected_null_hyp = rejected_null_hyp
        self.first_test_res = first_test_res

    def oracles(self):
        '''Final test (on list of tests, espacially in case of mulitple tests
        on common reference).

        :returns: list(bool)
        '''
        return [not np.any(rnh) for rnh in self.rejected_null_hyp]

    def __bool__(self):
        return not np.any(self.rejected_null_hyp)

    @property
    def nb_rejected(self):
        '''Number of rejected null hypothesis according Holm-Bonferroni test.

        :returns: list(int) or
            :class:`list` (:obj:`numpy.generic` (:obj:`int`))
        '''
        return [np.count_nonzero(rnh) for rnh in self.rejected_null_hyp]

    @property
    def rejected_proportion(self):
        '''Rejected proportion in percentage.

        :returns: :class:`list` (:class:`numpy.generic` (:obj:`float`))
        '''
        return [nr / rnh.size * 100
                for nr, rnh in zip(self.nb_rejected, self.rejected_null_hyp)]

    @property
    def sort_ordering(self):
        '''Get the sorted pvalues.

        :returns: :class:`list` (:class:`numpy.generic` (:obj:`int`))
        '''
        sort_inds = [np.argsort(pvals.flatten())
                     for pvals in self.first_test_res.pvalue]
        inv_sort = [np.argsort(si) for si in sort_inds]
        return [iv.reshape(pvals.shape)
                for pvals in self.first_test_res.pvalue
                for iv in inv_sort]


class TestHolmBonferroni(Test):
    '''Holm-Bonferroni method for multiple tests of the same hypothesis.'''

    def __init__(self, *, name, description='', labels=None, test, alpha=0.01):
        '''Initialisation of :class:`TestHolmBonferroni`.

        :param str name: local name of the test
        :param str description: specific description of the test
        :param dict labels: labels to be used for test classification in
                            reports (for example category, jdd name, type of
                            result, ...)
        :param test: test from which pvalues will be extracted
        :type test: :class:`valjean.gavroche.test.Test` child class
        :param float alpha: significance level

        The significance level is also considered two-sided here, so divide by
        2. This is the overall significance level.
        '''
        super().__init__(name=name, description=description, labels=labels)
        self.test = test
        self.alpha = alpha / 2

    @property
    def ntests(self):
        '''Returns the number of hypotheses to test, i.e. number of bins here.
        '''
        return self.test.dsref.size

    @staticmethod
    def holm_bonferroni_method(pvalues, alpha):
        '''Holm-Bonferroni method.

        :param pvalues: array of pvalues
        :type pvalues: :obj:`numpy.ndarray`
        :param float alpha: significance level chosen for the
                                         Holm-Bonferroni method
        :returns: sorted indices, array of the bins significance level, array
                  of rejection of the null hypothesis
        '''
        flat_pvals = pvalues.flatten()
        sorted_inds = np.argsort(flat_pvals)
        alpha_i, rejected_hyp = [], []
        for _i, pval in enumerate(flat_pvals[sorted_inds]):
            denom = flat_pvals.size - (_i+1) + 1
            alpha_i.append(alpha / denom)
            rejected_hyp.append(pval < alpha_i[-1])
        inv_sort = np.argsort(sorted_inds)
        return (np.array(alpha_i)[inv_sort].reshape(pvalues.shape),
                np.array(rejected_hyp)[inv_sort].reshape(pvalues.shape))

    def evaluate(self):
        '''Evaluation of the used test and of the Holm-Bonferroni method.

        :returns: :class:`~.TestResultHolmBonferroni`
        '''
        test_res = self.test.evaluate()
        hb_res = [self.holm_bonferroni_method(pvals, self.alpha)
                  for pvals in test_res.pvalue]
        return TestResultHolmBonferroni(self, test_res, *zip(*hb_res))

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        yield from self.test.data()
        yield float(self.alpha).hex().encode('utf-8')



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Module for the ``env`` subcommand.'''

import logging
from pprint import pprint

from ..common import Command
from ...cosette.env import Env


LOGGER = logging.getLogger(__name__)


class EnvCommand(Command):
    '''Command class for the ``env`` subcommand.'''

    NAME = 'env'

    HELP = 'Inspect the content of one or more serialized environment files.'

    DESC = HELP

    def register(self, parser):
        '''Register options for this command in the parser.'''
        parser.add_argument('--env-format', action='store',
                            choices=('pickle',), default='pickle',
                            help='environment persistency format')
        parser.add_argument('env_files', metavar='ENV_FILE', nargs='+',
                            help='The environment file to show.')
        parser.set_defaults(func=self.env)

    @staticmethod
    def env(args, _config):
        '''Execute the ``show-env`` command.'''
        # deserialize the environment
        for env_file in args.env_files:
            env = Env.from_file(env_file, fmt=args.env_format)
            if env is None:
                LOGGER.error('cannot show the environment for task %s',
                             env_file)
            else:
                pprint(dict(env.items()))



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module is designed to read all results contained an standard HDF5
output file from Apollo3 (with keff, fluxes and rates) and to store them in
memory as a :class:`~valjean.eponine.browser.Browser`. This makes it possible
to explore the content of the file using the facilities provided by
:class:`~valjean.eponine.browser.Browser`.

.. note::

    The main limitation is on speed, especially if the HDF5 file is big or
    contains a lot of nested groups. In that case, if you do not need to
    explore the file and you know exactly which results will be used, use
    :class:`~valjean.eponine.apollo3.hdf5_picker.Picker` instead.


Data models of Apollo3 HDF5 files
---------------------------------

The Apollo3 HDF5 output files mainly follow the standard data model, initially
designed for reaction rates, which is structured as follows:

.. code-block::

    ┌─ info ─┬─ (COMMENT)
    │        ├─ NOUT
    │        ├─ output_id1 ┬─ GEOMID
    │        │             ├─ NG
    │        │             └─ COMMENT
    │        ├─ output_id2 ┄
    │        ┆
    │        ┆
    │        └─ output_idNOUT ┄
    │
    ├─ geometry ─┬─ NGEO
    │            ├─ geometry_id1 ┬─ NZONE
    │            │               ├─ VOLUME[NZONE]
    │            │               └─ ZONENAME[NZONE]
    │            ├─ geometry_id2 ┄
    │            ┆
    │            ┆
    │            └─ geometry_idNGEO ┄
    │
    ├─ output_id1 ─┬─ totaloutput ─┬─ KEFF (float)
    │              │               ├─ (KINF (float))
    │              │               ├─ (ABSORPTION[NG])
    │              │               ├─ (CURRENT[NG])
    │              │               ├─ (FLUX[NG])
    │              │               ├─ (MIGRATIONAREA (float))
    │              │               ├─ (PRODUCTION[NG])
    │              │               ├─ (NVAL)
    │              │               ├─ (LOCALVALUE[NVAL])
    │              │               ├─ (LOCALNAME[NVAL])
    │              │               └─ (localvalue) ─┬─ LOCALNAME[NVAL]
    │              │                                ├─ value_id1
    │              │                                ├─ value_id2
    │              │                                ┆
    │              │                                ┆
    │              │                                └─ value_idNVAL
    │              ├─ zone_id_1 ─┬─ NISOT
    │              │             ├─ ISOTOPE[NISOT]
    │              │             ├─ CONCEN[NISOT]
    │              │             ├─ FLUX[NG]
    │              │             ├─ macro ─┬─ reaction_id[NG]
    │              │             │         ├─┄
    │              │             │         ┆
    │              │             │
    │              │             ├─ isotope_id1 ─┬─ reaction_id[NG]
    │              │             │               ├─┄
    │              │             │               ┆
    │              │             │               ┆
    │              │             │
    │              │             ├─ isotope_id2 ┄
    │              │             ┆
    │              │             ┆
    │              │             └─ isotope_idNISOT ┄
    │              ├─ zone_id2 ┄
    │              ┆
    │              ┆
    │              └─ zone_idNZONE ┄
    ├─ output_id2 ┄
    ┆
    ┆
    └─ output_idNOUT ┄

Elements in brackets are optional. If ``NISOT == 0``, the ``ISOTOPE``,
``CONCEN`` and ``isotope_id`` groups will be absent. The ``macro`` group
stores the macroscopic reaction rates (= reaction rates for a medium). The
``zone_id`` folder names appear as values in the ZONENAME dataset in the
geometry group, and they are arbitrary strings (chosen by the user in the
Apollo3 case). ``VOLUME`` stores the volumes (in cm³) of the zones, if needed
later for normalisation for example. Most physics quantities are given on
``NG`` energy groups but the group bounds are not stored in the file. Some
quantities are given as a function of energy groups and anisotropy. The order
of the anisotropy development is available in the ``'info'`` group, possibly in
the ``'isotope_id`` one. Current and surfacic flux (under ``'totaloutput'``)
have a ``surface id`` component in addition to energy groups.

Some files can also contain custom values or user values, especially in some
cases the standard structure does not appear at all (no ``zone_id`` or
``totaloutput`` folder) but directly local values:

.. code-block::

    ┌─ info ─┬─ (COMMENT)
    │        ├─ (FORMAT)
    │        ├─ (VERSION)
    │
    └─ output ─┬─ (LOCALVALUE[NVAL])
               ├─ (LOCALNAME[NVAL])
               └─ (localvalue) ─┬─ LOCALNAME[NVAL]
                                ├─ value_id1
                                ┆
                                ┆
                                └─ value_idNVAL


Use of :class:`Reader`
----------------------

.. code-block:: python

    from valjean.eponine.apollo3.hdf5_reader import Reader
    ap3r = Reader('MON_FICHIER.hdf')
    ap3b = ap3r.to_browser()

See :mod:`~valjean.eponine.browser` to get more details on how to use the
:class:`~valjean.eponine.browser.Browser` objects.

In the resulting :class:`~valjean.eponine.browser.Browser`, the metadata are
taken from the :obj:`h5py.Group` values, while the actual data are taken from
the :obj:`h5py.Dataset` ones.

Data are directly stored as :class:`~valjean.eponine.dataset.Dataset`. Errors
are set to ``numpy.nan`` by default but this can be changed and be set to 0
(for example) using the ``error_value`` argument to the constructor. Bins are
set to the group indices, or to group indices and number of anisotropies when
available. Two specific cases are added:
* surfacic flux: bins are given in group indices and surface numbers
* current: bins are given in group indices, surface numbers and direction

Some metadata keys in the browser are slightly different from the corresponding
names of the HDF5 file:

* in the global variables (``Browser.globals``):
    * under the ``'info'`` key:
        * ``'geom_id' `` for the ``'GEOMID'`` metadata
        * ``'ngroups'`` for ``'NG'``
    * under the ``'geometry'`` key: a dictonary associating the ``'geom_id'``
      names to dictionaries of {``'ZONENAME'``: ``'VOLUME'``}.
* in the results (``Browser.content``):
    * zone names: kept as they are given in the HDF5 file (also true for
      ``'totaloutput'``)
    * reaction names, flux, keff are set in lower case instead of upper
    * ``'CONCEN'`` is renamed in ``'concentration'`` and appears as a standard
      result (like reaction rates)
    * isotopes names are kept as they are but spaces surrounding the name are
      stripped. The ``'macro'`` key is kept unchanged for macroscopic results
      (integrated over isotopes)
    * local value names are left inchanged (under the ``'result_name'`` key).

The most common keys in the :class:`~valjean.eponine.browser.Browser` are
``['result_name', 'output', 'zone', 'isotope']``.


An inspect function is also available (no class:`Reader` object needed) to have
a quick look at the content of the file.

.. code-block:: python

    from valjean.eponine.apollo3.hdf5_reader import inspect
    inspect('MON_FICHIER.hdf')
'''

from collections import OrderedDict
import logging
import h5py
import numpy as np

from ...chrono import Chrono
from ..browser import Browser
from ..dataset import Dataset


LOGGER = logging.getLogger(__name__)


def hdf_to_browser(*args, **kwargs):
    '''Build a :class:`~valjean.eponine.browser.Browser` from an Apollo3 HDF5
    output.

    :param args: other arguments to be passed to :class:`Reader` constructor
    :param kargs: keyword arguments to be passed to :class:`Reader` constructor
    :returns: Browser
    '''
    return Reader(*args, **kwargs).to_browser()


class Reader:
    '''Read an HDF5 file from Apollo3.'''

    def __init__(self, fname, error_value=np.nan):
        '''Initialize the :class:`Reader` object and read the HDF5 file.

        :param str fname: path to the Apollo3 HDF5 ouput
        :param error_value: value to give to the error (default: ``numpy.nan``)
        :type error: int, float
        '''
        LOGGER.info('Reading %s', fname)
        self.error_value = error_value
        self.geom = {}
        self.info = {}
        self.res = {}
        with Chrono() as chrono:
            hfile = h5py.File(fname, 'r')
        LOGGER.info('hdf5 loading done in %f s', chrono)
        with chrono:
            self.read_file(hfile)
        hfile.close()
        LOGGER.info('Successful reading in %f s', chrono)

    def read_file(self, hfile):
        '''Read the HDF5 file.

        The information read depends on the type of file. In the case of a
        *standard* output file, we read ``'info'``, ``'geometry'`` and all
        results. In a *user* output, only results are read.

        :param str hfile: path to the HDF5 file

        :raises ReaderException: if unexpected keys are found.
        '''
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug('Keys in HDF5 file: %s', list(hfile.keys()))
        if 'info' not in hfile:
            raise ReaderException("'info' missing in the HDF5 file")
        self.info = extract_info(hfile['info'])
        if self.info:
            self.process_standard_values(hfile)
        else:
            self.process_user_values(hfile)

    def process_standard_values(self, hfile):
        '''Process standard values (keff, flux, rates).'''
        self.geom = extract_geometry(hfile['geometry'])
        for key, value in hfile.items():
            if not key.startswith('output_'):
                continue
            self.res[key] = extract_standard_values(
                value, self.info[key]['ngroups'], self.error_value)

    def process_user_values(self, hfile):
        '''Process user values stored in files (named local values in file).
        '''
        for key, value in hfile.items():
            if key == 'info':
                continue
            self.res[key] = extract_user_values(value, self.error_value)

    def to_browser(self):
        '''Build a :class:`~valjean.eponine.browser.Browser` from
        the results in the HDF5 file.

        :rtype: Browser
        '''
        browser = Browser(
            dict_to_list('output', self.res),
            global_vars={'info': self.info, 'geometry': self.geom})
        return browser


def extract_geometry(geometry):
    '''Extract geometry in dict to go to
    :class:`~valjean.eponine.browser.Browser` globals.

    :param h5py.Group geometry: geometry stored in HDF5
    '''
    LOGGER.debug("In extract_geometry")
    dgeom = {}
    for geom in geometry:
        if geom == 'NGEO':
            continue
        igeom = geometry[geom]
        dgeom[geom] = {
            name.decode('UTF-8').strip(): vol
            for name, vol in zip(igeom['ZONENAME'], igeom['VOLUME'])}
    LOGGER.debug("geometry dict done")
    return dgeom


def extract_info(info):
    '''Extract info in dict to go to
    :class:`~valjean.eponine.browser.Browser` globals.

    :param h5py.Group info: info stored in HDF5
    '''
    LOGGER.debug("In extract_info")
    dinfo = {}
    if 'NOUT' not in info:
        LOGGER.debug('No description of output* folders, info not stored')
        LOGGER.debug('keys are %s', list(info.keys()))
        return dinfo
    infos = [(k, v) for k, v in info.items() if k.startswith('output_')]
    for nout, a_info in infos:
        dinfo[nout] = {
            'geom_id': a_info['GEOMID'][...][0].decode('UTF-8'),
            'ngroups': a_info['NG'][...][0]}
    LOGGER.debug("Leaving extract_info")
    return dinfo


def extract_standard_values(val, ngroups, error):
    '''Extract results from Apollo3 HDF5 file.

    :param str out_id: name of the output folder in HDF5
    :param h5py.Group val: results stored in the folder
    :param error: value to give to the error (default: ``numpy.nan``)
    :type error: int, float
    :returns: list of all results stored with metadata and data shaped as
        `Dataset`
    :rtype: list(dict)
    '''
    tdict = {}
    for zone in val:
        if zone == 'totaloutput':
            tdict[zone] = loop_over_std_values(
                val['totaloutput'], ngroups, error)
        else:
            tdict[zone] = extract_zone_values(val[zone], ngroups, error)
    return dict_to_list('zone', tdict)


def extract_user_values(val, error):
    '''Extract user results (or 'local' values).

    :param h5py.Group val: results stored under ``'LOCALVALUES'``
    :param error: value to give to the error (default: ``numpy.nan``)
    :type error: int, float
    :returns: list of results including metadata
    :rtype: list(dict)
    '''
    if 'LOCALNAME' in val:
        return extract_localvalues(val['LOCALNAME'], val['LOCALVALUE'], error)
    rlist = []
    locval = 'localvalue'
    lkeys = [e.decode('UTF-8').strip()
             for e in val[locval]['LOCALNAME'][...]]
    rlist = [{'result_name': k,
              'results': hdfdataset_to_dataset(val[locval][k], k, error)}
             for k in lkeys]
    return rlist


def extract_localvalues(lnames, lvals, error):
    '''Extract local values from a list of local names.

    :param list(str) lnames: list of the names of the local values
    :param h5py.Dataset lvals: local (scalar) values
    :param error: value to give to the error (default: ``numpy.nan``)
    :type error: int, float
    :returns: list of results including metadata
    :rtype: list(dict)
    '''
    LOGGER.debug('In extract_localvalues')
    return [{'result_name': key.decode('UTF-8').strip(),
             'results': build_dataset(val, error, key.decode('UTF-8').strip())}
            for key, val in zip(lnames, lvals)]


def loop_over_std_values(lres, ngroups, error):
    '''Loop over results stored in the ``'output_*'`` groups of the HDF5
    file.

    Also retrieve number of anisotropies from the `info` :obj:`h5py.Group`.

    :param h5py.Group lres: list of results
    :param int ngroups: number of energy groups
    :param error: value to give to the error (default: ``numpy.nan``)
    :type error: int, float
    :returns: list of results including metadata
    :rtype: list(dict)
    '''
    rlist = []
    res = [(k, v) for k, v in lres.items()
           if not k.startswith(('LOCAL', 'local', 'info', 'NSURF'))]
    anisotropies = extract_output_info(lres.get('info'))
    nsurfaces = extract_surfaces_number(lres.get('NSURF'))
    for nres, vres in res:
        bins = make_bins(nres, vres, ngroups, anisotropies, nsurfaces)
        rlist.append({
            'result_name': nres.lower(),
            'results': hdfdataset_to_dataset(
                vres, what=nres.lower(), error=error, bins=bins)})
    if 'LOCALNAME' in lres or 'localvalue' in lres:
        rlist.extend(extract_user_values(lres, error))
    return rlist


def hdfdataset_to_dataset(hdf_data, what, error, bins=None):
    '''Build :class:`~valjean.eponine.dataset.Dataset` from HDF5 dataset (in
    *rates* cases).

    Returns a scalar or an array depending on data shape.

    :param h5py.Dataset hdf_data: dataset containing the data
    :param str what: quantity name
    :param error: value to give to the error (default: ``numpy.nan``)
    :type error: int, float
    :param collections.OrderedDict bins: bins corresponding to data, default
      `None`
    :rtype: Dataset
    '''
    if not isinstance(hdf_data, h5py.Dataset):
        raise TypeError(f'Expected a h5py.Dataset, got a {type(hdf_data)}')
    if hdf_data.shape == (1,) and not bins:  # ngroups == 0:
        val = hdf_data[...][0]
    else:
        val = hdf_data[...]
    return build_dataset(val, error, what, bins)  # ngroups, naniso)


def build_dataset(data, error, what, bins=None):
    '''Build :class:`~valjean.eponine.dataset.Dataset` from HDF5 result.

    :param data: data
    :type data: numpy.generic, numpy.ndarray
    :param str what: quantity name
    :param error: value of error
    :type error: numpy.generic, float, int, default = ``numpy.nan``
    :param str what: quantity name
    :param collections.OrderedDict bins: bins corresponding to data, default
      `None`
    :rtype: Dataset
    '''
    error = np.full_like(data, error,
                         dtype=(float if np.isnan(error) else data.dtype))
    if not bins:
        return Dataset(value=data, error=error, what=what)
    if 'anisotropies' in bins:
        ngroups = len(bins['groups'])
        naniso = len(bins['anisotropies'])
        return Dataset(value=data.reshape(naniso, ngroups),
                       error=error.reshape(naniso, ngroups),
                       what=what, bins=bins)
    if 'incident neutron groups' in bins:
        ngroups = len(bins['groups'])
        nincg = len(bins['incident neutron groups'])
        return Dataset(value=data.reshape(nincg, ngroups),
                       error=error.reshape(nincg, ngroups),
                       what=what, bins=bins)
    return Dataset(value=data, error=error, what=what, bins=bins)


def make_bins(nres, vres, ngroups, anisotropies=None, nsurfaces=None):
    '''Build bins :obj:`collections.OrderedDict`.

    The bins currently available are
    * number of energy groups
    * number of anisotropies (for isotopes reaction rates)
    * number of surfaces (for current and surface flux)

    :param h5py.Group data: `info` block from `output_*`
    :rtype: dict
    :returns: dictionary of all possible number of bins
    '''
    if nres in ('KEFF', 'KINF', 'MIGRATIONAREA'):
        return None
    data = vres[...]
    if ngroups == data.size:
        return OrderedDict([('groups', np.arange(ngroups))])
    if nres == 'SURFFLUX':
        return OrderedDict([('groups', np.arange(ngroups)),
                            ('surfaces', np.arange(nsurfaces))])
    if nres == 'CURRENT':
        return OrderedDict([('groups', np.arange(ngroups)),
                            ('surfaces', np.arange(nsurfaces)),
                            ('direction', np.array(['incoming', 'leaving']))])
    if nres == 'MultigroupSpectrum':
        incgrps = data.size // ngroups
        return OrderedDict([('incident neutron groups', np.arange(incgrps)),
                            ('groups', np.arange(ngroups))])
    naniso = anisotropies.get(nres, anisotropies['anisotropy'])
    if ngroups * naniso != data.size:
        raise ReaderException(
            f"Dataset size ({data.size})not corresponding to "
            f"NG*Nanisotropies {ngroups}*{naniso} for {nres}")
    return OrderedDict([('anisotropies', np.arange(naniso)),
                        ('groups', np.arange(ngroups))])


def extract_output_info(data, name='anisotropy'):
    '''Extract info from output group containing number of anisotripies.

    :param h5py.Group data: `info` block from `output_*`
    :param str name: anisotropy name (`'anisotropy'` per default, result name
      if one
    :rtype: dict
    :returns: dictionary indexed by anisotropy name
    '''
    aniso = {'anisotropy': 1}
    if not data:
        return aniso
    for dname, dat in data.items():
        if 'macro' in dat.name and isinstance(dat, h5py.Group):
            aniso.update(extract_output_info(dat, name=dname))
            continue
        if dname != 'nbAnisotropy':
            raise ReaderException('info group should contain nbAnisotropy')
        aniso[name] = dat[...][0]
    return aniso


def extract_surfaces_number(data):
    '''Extract number of surfaces if available.'''
    if not data:
        return None
    return data[...][0]


def extract_concentrations(zone, error):
    '''Store isotopes list and concentration from the zone group.

    :param h5py.Group zone: HDF5 group corresponding to the zone
    :param error: value to give to the error (default: ``numpy.nan``)
    :type error: int, float
    :returns: list of dictionaries with isotope name in metadata,
        ``'concentration'`` as ``result_name`` and concentration values
        stored in a :class:`~valjean.eponine.dataset.Dataset` under the
        ``'results'`` key
    :rtype: list(dict)
    '''
    tdict = OrderedDict()
    for isotop, conc in zip(zone.get('ISOTOPE'), zone.get('CONCEN')):
        tdict[isotop.decode('UTF-8').strip()] = {
            'result_name': 'concentration',
            'results': Dataset(value=conc,
                               error=np.full_like(conc, error, dtype=float),
                               what='concentration')}
    return dict_to_list('isotope', tdict)


def extract_zone_values(val, ngroups, error):
    '''Store results by zone.

    Results are stored depending on their nature (flux, concentrations,
    etc.).

    :param h5py.Group val: HDF5 group containing results from a given zone
    :returns: list of results including metadata
    :param error: value to give to the error (default: ``numpy.nan``)
    :type error: int, float
    :rtype: list(dict)
    '''
    tlist = []
    isotd = {}
    if val['NISOT'][...][0] != 0:
        liso = set(iso.decode('UTF-8').strip() for iso in val['ISOTOPE'])
    else:
        liso = {}
    for key in val:
        if key == 'CONCEN':
            tlist.extend(extract_concentrations(val, error))
        elif key == 'macro':
            isotd[key] = loop_over_std_values(val['macro'], ngroups, error)
        elif key in ('NISOT', 'ISOTOPE'):
            continue
        elif key in liso:
            isotd[key] = loop_over_std_values(val[key], ngroups, error)
        else:
            bins = make_bins(key, val[key], ngroups)
            tlist.append({'result_name': key.lower(),
                          'results': hdfdataset_to_dataset(
                              val[key], key.lower(), error, bins)})
    tlist.extend(dict_to_list('isotope', isotd))
    return tlist


def dict_to_list(key_name, tdict):
    '''Transform a dictionary in a list of dictionaries including the key
    as a new key, value member of each dictionary.

    :param str key_name: key of the new dictionary
    :param dict tdict: dictionary whose keys will become values of key_name
    :returns: list of results including metadata
    :rtype: list(dict)

    >>> tdict = OrderedDict()
    >>> tdict['Graham'] = [{'day': 'Monday', 'meal': 'spam'},
    ...                    {'day': 'Tuesday', 'meal': 'egg'}]
    >>> tdict['Terry'] =  {'day': 'Monday', 'meal': 'bacon'}
    >>> tlist = dict_to_list('consumer', tdict)
    >>> len(tlist)
    3
    >>> from pprint import pprint
    >>> pprint(tlist)
    [{'consumer': 'Graham', 'day': 'Monday', 'meal': 'spam'},
     {'consumer': 'Graham', 'day': 'Tuesday', 'meal': 'egg'},
     {'consumer': 'Terry', 'day': 'Monday', 'meal': 'bacon'}]

    .. todo::

        Simplify this example by changing back OrderedDict in dict when python
        3.5 won't be anymore supported.
    '''
    tlist = []
    for k, val in tdict.items():
        if isinstance(val, dict):
            val_copy = val.copy()
            val_copy[key_name] = k
            tlist.append(val_copy)
        elif isinstance(val, list):
            for dic in val:
                dic_copy = dic.copy()
                dic_copy[key_name] = k
                tlist.append(dic_copy)
    return tlist


def inspect(hfile, name_spaces=50):
    '''Loop recursively over content of the file and print it.

    :param str hfile: path to the Apollo3 HDF5 ouput
    :param int name_spaces: number of spaces reserved for the folder's names,
        default = 50
    '''
    def find_subobj(name, obj):
        LOGGER.info(f'{name:<{name_spaces}}  {obj}')  # pylint: disable=W1203
    hfile.visititems(find_subobj)


class ReaderException(Exception):
    '''An error that may be raised by the :class:`Reader` class.'''



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

''':class:`Config` objects encapsulate a set of configuration options for a
:mod:`valjean` run. Here is how you create one:

    >>> from valjean.config import Config
    >>> config = Config()

By default, :class:`Config` objects come with a ``'path'`` configuration
section, which may be used to set default values for any configuration option.
A few options are set from the beginning:

    >>> for opt, val in sorted(config['path'].items()):
    ...     print(f'{opt} = {val}')
    log-root = /.../log
    output-root = /.../output
    report-root = /.../report

The :class:`Config` class behaves like a simple dictionary:

    >>> print(config.query('path', 'report-root'))
    /.../report

Module API
----------
'''

from collections.abc import MutableMapping
from pathlib import Path
import toml


class Config(MutableMapping):
    '''The base configuration class for :mod:`valjean`.'''

    @classmethod
    def from_file(cls, path):
        '''Construct a configuration object from a TOML file.

        :param path: A path for the configuration file.
        :type path: pathlib.Path or str
        '''
        file_content = toml.load(str(path))
        return cls(file_content)

    def __init__(self, dictionary=None):
        '''Construct a configuration object from a dictionary.

        The configuration will be initialized to contain a few default options.

        :param dict dictionary: The configuration object.
        :returns: The constructed Config object.
        '''
        self.conf = dict(dictionary) if dictionary is not None else {}

        # sanity check
        if 'path' in self.conf and not isinstance(self.conf['path'], dict):
            raise ValueError('the "path" option is forbidden at the top level')

        # Set some default options
        if 'path' not in self:
            self['path'] = {}
        conf_path = self['path']
        work_dir = Path.cwd()
        if 'log-root' not in conf_path:
            conf_path['log-root'] = f'{work_dir}/log'
        if 'output-root' not in conf_path:
            conf_path['output-root'] = f'{work_dir}/output'
        if 'report-root' not in conf_path:
            conf_path['report-root'] = f'{work_dir}/report'

    def __getitem__(self, key):
        return self.conf[key]

    def __setitem__(self, key, value):
        self.conf[key] = value

    def __delitem__(self, key):
        del self.conf[key]

    def __iter__(self):
        yield from self.conf

    def __len__(self):
        return len(self.conf)

    def __eq__(self, other):
        if not isinstance(other, Config):
            return False
        return self.conf == other.conf

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return toml.dumps(self.conf)

    def __repr__(self):
        return f'Config({self.conf})'

    def query(self, section, option):
        '''Return the value of `option` from `section`.'''
        return self.conf[section][option]

    def set(self, section, option, value):
        '''Set the value of `option` in `section` to be `value`.'''
        self.conf[section][option] = value



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
# -*- coding: utf-8 -*-
"""This submodule contains a few useful tasks for checking out, configuring,
and building arbitrary code.

The :class:`CheckoutTask` task class checks out a version-controlled
repository. For the moment, only ``git`` repositories are supported. The path
to the ``git`` executable may be specified through the :data:`CheckoutTask.GIT`
class variable.

.. todo::

   Implement ``svn`` and ``cvs`` checkout; ``copy`` checkout (i.e. copy a
   directory from somewhere) may also be useful.

The :class:`BuildTask` task class builds code from a given source directory. A
build directory must also be specified and will be created if necessary. For
the moment, :class:`BuildTask` only supports ` ``cmake`` builds, but there are
plans to add support for ``autoconf``/``configure``/``make`` builds. The path
to the ``cmake`` executable may be specified through the
:data:`BuildTask.CMAKE` class variable.

.. todo::

   Implement ``autoconf``/``configure``/``make`` builds.

.. doctest:: code
   :hide:

   >>> from valjean.cosette.code import CheckoutTask, BuildTask
   >>> import subprocess
   >>> work_dir = getfixture('tmp_path')
   >>> repo_dir = work_dir / 'repo'
   >>> repo_dir.mkdir()
   >>> subprocess.check_call([CheckoutTask.GIT, 'init', str(repo_dir)])
   0
   >>> cmakelists_path = repo_dir / 'CMakeLists.txt'
   >>> with cmakelists_path.open('w') as cmake_file:
   ...     print('project(TestCodeTasks C)\\n'
   ...           'cmake_minimum_required(VERSION 2.6)\\n'
   ...           'set(SOURCE_FILENAME "${PROJECT_BINARY_DIR}/test.c")\\n'
   ...           'file(WRITE "${SOURCE_FILENAME}" "int main(){return 0;}")\\n'
   ...           'add_executable(test_exe "${SOURCE_FILENAME}")\\n',
   ...           file=cmake_file)
   >>> git_dir = repo_dir / '.git'
   >>> subprocess.check_call([CheckoutTask.GIT, '--git-dir', str(git_dir),
   ...                        'config', 'user.email', 'sblinda@antani.com'])
   0
   >>> subprocess.check_call([CheckoutTask.GIT, '--git-dir', str(git_dir),
   ...                        'config', 'user.name', 'Conte Mascetti'])
   0
   >>> subprocess.check_call([CheckoutTask.GIT, '--git-dir', str(git_dir),
   ...                        '--work-tree' , str(repo_dir),
   ...                        'add', 'CMakeLists.txt'])
   0
   >>> subprocess.check_call([CheckoutTask.GIT, '--git-dir', str(git_dir),
   ...                        '--work-tree', str(repo_dir),
   ...                        'commit', '-a', '-m', 'Test commit'])
   0

To describe the usage of :class:`CheckoutTask` and :class:`BuildTask`, let us
assume that ``repo_dir`` contains a ``git`` repository with a CMake project.
We use a temporary directory ``work_dir`` for our test:

   >>> checkout_dir = work_dir / 'checkout'
   >>> build_dir = work_dir / 'build'
   >>> log_dir = work_dir / 'log'

Now we can build checkout and build tasks for this repository:

   >>> from valjean.cosette.code import CheckoutTask, BuildTask
   >>> from pprint import pprint
   >>> ct = CheckoutTask(name='project_checkout',
   ...                   repository=repo_dir,
   ...                   checkout_root=checkout_dir,
   ...                   log_root=log_dir)
   >>> bt = BuildTask(name='project_build',
   ...                source=ct,
   ...                build_root=build_dir,
   ...                build_flags=['--' ,'-j4'],
   ...                log_root=log_dir)

Note how we passed the `ct` object directly to the `source` argument of the
:class:`BuildTask` constructor: we are telling the :class:`BuildTask` to look
for the sources to build in the checkout directory. You can also pass a normal
path to the `source` argument instead.

   >>> from valjean.cosette.env import Env
   >>> env = Env()
   >>> ct_up, ct_status = ct.do(env, config=None)
   >>> print(ct_status)
   TaskStatus.DONE
   >>> pprint(ct_up)
   {'project_checkout': {'checkout_log': '.../log/project_checkout.log',
                         'elapsed_time': ...,
                         'output_dir': '.../checkout/project_checkout',
                         'repository': '.../repo'}}
   >>> env.apply(ct_up)  # apply CheckoutTask's environment update
   ...                   # for this example, this is actually optional
   >>> bt_up, bt_status = bt.do(env=env, config=None)
   >>> print(bt_status)
   TaskStatus.DONE
   >>> pprint(bt_up)
   {'project_build': {'build_log': '.../log/project_build.log',
                      'elapsed_time': ...,
                      'output_dir': '.../build/project_build'}}
"""

import os
import logging

from ..path import ensure
from ..chrono import Chrono
from .run import run
from .task import TaskStatus
from .pythontask import PythonTask


LOGGER = logging.getLogger(__name__)


class CheckoutTask(PythonTask):
    '''Task to check out code from a version-control system.  The actual code
    checkout is performed when the task is executed.
    '''

    def __init__(self, name, *, repository, checkout_root=None, log_root=None,
                 flags=None, ref=None, vcs='git', deps=None, soft_deps=None):
        # pylint: disable=too-many-arguments
        '''Construct a :class:`CheckoutTask`.

        :param str name: The name of this task.
        :param str checkout_root: The directory where the code will be checked
                                  out, or `None` for the configuration default.
        :param str repository: The repository for checkout.
        :param str log_root: The path to the log directory, or `None` for the
                             configuration default.
        :param flags: The flags to be used at checkout time, as a list of
                      strings.
        :type flags: list(str) or None
        :param ref: The reference to check out.
        :type ref: str or None
        :param vcs: The version-control system to use. Must be one of:
                    ``'git'`` (default), ``'svn'``, ``'cvs'``, ``'copy'``.
        :type vcs: str or None
        :param deps: The dependencies for this task (see
               :meth:`Task.__init__() <valjean.cosette.task.Task.__init__>`
               for the format), or `None`.
        :type deps: list(Task) or None
        :param soft_deps: The soft dependencies for this task (see
               :meth:`Task.__init__() <valjean.cosette.task.Task.__init__>`
               for the format), or `None`.
        :type soft_deps: list(Task) or None
        '''
        self.checkout_root = checkout_root
        self.log_root = log_root

        if vcs == 'git':
            if ref is None:
                ref = 'master'

            def checkout_vcs(checkout_dir, log):
                clone_cli = [self.GIT, 'clone']
                if flags is not None:
                    clone_cli.extend(flags)
                clone_cli.extend(['--', str(repository), str(checkout_dir)])
                ret, status, _ = run([clone_cli], stdout=log, stderr=log)
                if ret[-1] != 0:
                    LOGGER.debug('`git clone` returned %s', ret)
                    return status
                checkout_cli = [self.GIT, 'checkout', ref]
                ret, status, _ = run([checkout_cli], stdout=log, stderr=log,
                                     cwd=str(checkout_dir))
                if ret[-1] != 0:
                    LOGGER.debug('`git checkout` returned %s', ret)
                return status
        elif vcs == 'svn':
            raise NotImplementedError('SVN checkout not implemented yet')
        elif vcs == 'cvs':
            raise NotImplementedError('CVS checkout not implemented yet')
        elif vcs == 'copy':
            raise NotImplementedError('copy checkout not implemented yet')
        else:
            raise ValueError(f'unrecognized VCS: {vcs}')

        def checkout(*, config):
            # setup log dir and file
            if self.log_root is None:
                self.log_root = config.query('path', 'log-root')
            log_file = ensure(self.log_root, self.name + '.log')
            ensure(log_file)

            # setup checkout dir
            if self.checkout_root is None:
                self.checkout_root = config.query('path', 'output-root')
            checkout_dir = ensure(self.checkout_root, self.name, is_dir=True)

            with log_file.open('w', encoding='utf-8') as log:
                with Chrono() as chrono:
                    status = checkout_vcs(checkout_dir, log)

            if status != TaskStatus.DONE:
                LOGGER.warning('CheckoutTask %s did not succeed (status: %s)',
                               self.name, status)
                if LOGGER.isEnabledFor(logging.DEBUG):
                    with log_file.open(encoding='utf-8') as log:
                        LOGGER.debug('checkout log:\n%s', log.read())

            env_up = {self.name: {'checkout_log': str(log_file),
                                  'output_dir': str(checkout_dir),
                                  'repository': str(repository),
                                  'elapsed_time': float(chrono)}}
            return env_up, status

        super().__init__(name, checkout, deps=deps, soft_deps=soft_deps,
                         config_kwarg='config')

        LOGGER.debug('Created %s task %r', self.__class__.__name__, self.name)

    #: Path to the :file:`git` executable. May be overridden before class
    #: instantiation.
    GIT = 'git'


class BuildTask(PythonTask):
    '''Task to build an existing source tree. The build is actually performed
    when the task is executed.
    '''

    def __init__(self, name, source, *, build_root=None, log_root=None,
                 targets=None, build_system='cmake', configure_flags=None,
                 build_flags=None, deps=None, soft_deps=None):
        # pylint: disable=too-many-arguments
        '''Construct a :class:`BuildTask`.

        :param str name: The name of this task.
        :param str source: The path to the directory containing the sources, or
                           a :class:`CheckoutTask` object (in which case the
                           checkout directory will be assumed to contain the
                           sources).
        :param str build_root: The path to the build directory (a subdirectory
                               will be created).
        :param str log_root: The path to the log directory.
        :param targets: A list of targets to build, or `None` for the default
                        target.
        :type targets: list(str) or None
        :param str build_system: The name of the build system to use. Must be
                                 one of: ``'cmake'`` (default),
                                 ``'configure'``.
        :param configure_flags: The flags that will be passed to the build tool
                                at configuration time, as a list of strings.
        :type configure_flags: list
        :param build_flags: The flags that will be passed to the build tool at
                            build time, as a list of strings.
        :type build_flags: list
        :param deps: The dependencies for this task (see
                     :meth:`Task.__init__()
                     <valjean.cosette.task.Task.__init__>` for the type), or
                     `None`.
        :type deps: list(Task) or None
        :param soft_deps: The soft dependencies for this task (see
                          :meth:`Task.__init__()
                          <valjean.cosette.task.Task.__init__>` for the type),
                          or `None`.
        :type soft_deps: list(Task) or None
        '''
        assert isinstance(source, (str, CheckoutTask))
        LOGGER.debug('BuildTask %s will look for source files in %s',
                     name, source)

        if deps is None:
            deps = []
        if isinstance(source, CheckoutTask):
            deps.append(source)

        self.log_root = log_root
        self.build_root = build_root

        if build_system == 'cmake':
            build_sys = self.cmake_build_sys(targets, configure_flags,
                                             build_flags)
        elif build_system in ('autoconf', 'configure'):
            raise NotImplementedError('configure build not implemented yet')
        else:
            raise ValueError(f'unrecognized build system: {build_system}')

        def build(*, config, env):
            # setup log dir and files
            if self.log_root is None:
                self.log_root = config.query('path', 'log-root')
            log_file = ensure(self.log_root, self.name + '.log')
            ensure(log_file)

            # setup build dir
            if self.build_root is None:
                self.build_root = config.query('path', 'output-root')
            build_dir = (ensure(self.build_root, self.name, is_dir=True)
                         .resolve())

            if isinstance(source, str):
                source_dir = os.path.abspath(source)
            else:
                source_dir = os.path.abspath(env[source.name]['output_dir'])

            with Chrono() as chrono:
                with log_file.open('w', encoding='utf-8') as log:
                    status = build_sys(source_dir, build_dir, log)

            if status != TaskStatus.DONE:
                LOGGER.warning('BuildTask %s did not succeed (status: %s)',
                               self.name, status)
                if LOGGER.isEnabledFor(logging.DEBUG):
                    with log_file.open(encoding='utf-8') as log:
                        LOGGER.debug('build log:\n%s', log.read())

            env_up = {self.name: {'build_log': str(log_file),
                                  'output_dir': str(build_dir),
                                  'elapsed_time': float(chrono)}}
            return env_up, status

        super().__init__(name, build, deps=deps, soft_deps=soft_deps,
                         env_kwarg='env', config_kwarg='config')

        LOGGER.debug('Created %s task %r', self.__class__.__name__, self.name)

    def cmake_build_sys(self, targets, configure_flags, build_flags):
        '''Return a function that builds a project using cmake.'''

        def build_sys(source_dir, build_dir, log):
            '''Build a project using cmake.'''
            configure_cli = [self.CMAKE]
            if configure_flags is not None:
                configure_cli.extend(configure_flags)
            configure_cli.append(source_dir)
            ret, status, _ = run([configure_cli], stdout=log, stderr=log,
                                 cwd=str(build_dir))
            if ret[-1] != 0:
                LOGGER.debug('`cmake` (configure step) returned %s', ret)
                return status
            build_cli = [self.CMAKE, '--build', str(build_dir)]
            target_list = [] if targets is None else targets
            for target in target_list:
                build_cli.append('--target')
                build_cli.append(target)
            if build_flags is not None:
                build_cli.extend(build_flags)
            ret, status, _ = run([build_cli], stdout=log, stderr=log,
                                 cwd=str(build_dir))
            if ret[-1] != 0:
                LOGGER.debug('`cmake` (build step) returned %s', ret)
            return status
        return build_sys

    #: Path to the :file:`cmake` executable. May be overridden before class
    #: instantiation.
    CMAKE = 'cmake'
